# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantization calibration methods."""

import abc
from typing import Union
import flax.struct
import jax.numpy as jnp


class Calibration(abc.ABC):

    @abc.abstractmethod
    def get_bound(self, x, shared_axes) -> jnp.ndarray:
        pass


@flax.struct.dataclass
class ConstantCalibration(Calibration):
    bound: Union[jnp.ndarray, float]

    def get_bound(self, x, shared_axes) -> jnp.ndarray:
        """Calibration."""
        del shared_axes
        assert self.bound > 0, 'Bound should be positive.'
        return jnp.asarray(self.bound).reshape((1,) * len(x.shape))


@flax.struct.dataclass
class AbsMaxCalibration(Calibration):
    """Simple max(abs(x)) calibration."""

    def get_bound(self, x, shared_axes) -> jnp.ndarray:
        """Calibration."""
        msg = 'Perhaps you are using fake_quant and forgot to set them.'
        assert shared_axes is not None, msg

        # NOTE: If you want to clip, consider using clip and clip_gradient in
        # int_numerics.IntNumerics.
        abs_max = jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True)
        abs_max = jnp.where(abs_max == 0.0, jnp.ones_like(abs_max), abs_max)
        return abs_max



import flax.linen as nn
import jax
import jax.lax as lax
import jax.numpy as jnp
from einops import rearrange
from functools import partial
from .flash_attention_tpu import _flash_attention_fwd, _flash_attention_bwd, BlockSizes
from .flash_attention_gpu import _mha_forward, _mha_backward


def _ring_attention_fwd(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    numerator = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(q.dtype)
    denominator = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
    axis_size = lax.psum(1, axis_name)
    block_size = q_len  # assumes this function is pre-sharded inside shard_map
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    def scan_kv_block(carry, idx):
        prev_max_score, numerator, denominator, k, v = carry
        ''' einsum version
        mask = lax.dynamic_slice_in_dim(attn_mask,
            (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        attn_weights = jnp.einsum("bqhd,bkd->bhqk", q, k) / scale
        attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        max_score = jnp.maximum(prev_max_score, jnp.max(attn_weights, axis=-1))
        exp_weights = jnp.exp(attn_weights - max_score[..., None])
        correction = rearrange(jnp.exp(prev_max_score - max_score), 'b h q -> b q h')[..., None]
        numerator = numerator * correction + jnp.einsum("bhqk,bkd->bqhd", exp_weights, v)
        denominator = denominator * jnp.exp(prev_max_score - max_score) + jnp.sum(exp_weights, axis=-1)
        '''
        attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
                                                   (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len,
                                                   axis=-1
                                                   )
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        numerator, denominator, max_score = _blockwise_attention_fwd(q, k, v, (numerator, denominator, prev_max_score),
                                                                     q_chunk_idx_start, k_chunk_idx_start,
                                                                     bias=attn_bias_slice, **blockwise_kwargs)
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]),
                   (k, v))
        return (max_score, numerator, denominator, k, v), None

    prev_max_score = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)
    (max_score, numerator, denominator, _, _), _ = lax.scan(scan_kv_block,
                                                            init=(prev_max_score, numerator, denominator, k, v),
                                                            xs=jnp.arange(0, axis_size))
    output = numerator / rearrange(denominator, 'b h q -> b q h')[..., None]
    return output.astype(v.dtype), (output, q, k, v, attn_bias, denominator, max_score)


def _ring_attention_bwd(axis_name, float32_logits, blockwise_kwargs, res, g):
    del float32_logits
    output, q, k, v, attn_bias, denominator, max_score = res
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    axis_size = lax.psum(1, axis_name)
    dq = jnp.zeros_like(q, dtype=q.dtype)
    dk = jnp.zeros_like(k, dtype=k.dtype)
    dv = jnp.zeros_like(v, dtype=k.dtype)
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]
    block_size = q.shape[1]  # assumes this function is pre-sharded inside shard_map

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        ''' einsum version
        # mask = lax.dynamic_slice_in_dim(attn_mask,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        # attn_weights = jnp.einsum("bqhd,bkd->bhqk", q, k) / scale
        # attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        # exp_weights = jnp.exp(attn_weights - max_score[..., None]) / denominator[..., None]
        # ds = jnp.einsum("bqhd,bkd->bhqk", g, v)
        # dl = (ds - jnp.einsum("bqhd,bqhd->bhs", g, output)[..., None]) * exp_weights
        # dq = dq + jnp.einsum("bhqk,bkd->bqhd", dl, k) / scale
        # dk = dk + jnp.einsum("bqhd,bhqk->bkd", q, dl) / scale
        # dv = dv + jnp.einsum("bhqk,bqhd->bkd", exp_weights, g)
        '''
        attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
                                                   (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len,
                                                   axis=-1
                                                   )
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        dq, dk, dv = _blockwise_attention_bwd(q, k, v, g, (dq, dk, dv, output, denominator, max_score),
                                              q_chunk_idx_start, k_chunk_idx_start, bias=attn_bias_slice,
                                              **blockwise_kwargs)
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(k.dtype)
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6])
def ring_attention(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    y, _ = _ring_attention_fwd(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs)
    return y


ring_attention.defvjp(_ring_attention_fwd, _ring_attention_bwd)


def _ring_attention_standard_fwd(q, k, v, attn_mask, axis_name, float32_logits):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    batch, q_len, num_heads, _ = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    numerator = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(q.dtype)
    denominator = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
    axis_size = lax.psum(1, axis_name)
    scale = jnp.sqrt(q.shape[-1])

    def scan_kv_block(carry, idx):
        prev_max_score, numerator, denominator, k, v = carry
        mask = lax.dynamic_slice_in_dim(attn_mask,
                                        (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        attn_weights = jnp.einsum("bqhd,bkhd->bhqk", q, k) / scale
        # attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        attn_weights = jnp.where(mask, attn_weights, -jnp.inf)
        max_score = jnp.maximum(prev_max_score, jnp.max(attn_weights, axis=-1))
        exp_weights = jnp.exp(attn_weights - max_score[..., None])
        correction = rearrange(jnp.exp(prev_max_score - max_score), 'b h q -> b q h')[..., None]
        numerator = numerator * correction + jnp.einsum("bhqk,bkhd->bqhd", exp_weights, v)
        denominator = denominator * jnp.exp(prev_max_score - max_score) + jnp.sum(exp_weights, axis=-1)
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                               (i + 1) % axis_size) for i in range(axis_size)]), (k, v))
        return (max_score, numerator, denominator, k, v), None

    prev_max_score = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)
    (max_score, numerator, denominator, _, _), _ = lax.scan(scan_kv_block,
                                                            init=(prev_max_score, numerator, denominator, k, v),
                                                            xs=jnp.arange(0, axis_size))
    output = numerator / rearrange(denominator, 'b h q -> b q h')[..., None]
    return output.astype(v.dtype), (output, q, k, v, attn_mask, numerator, denominator, max_score)


def _ring_attention_standard_bwd(axis_name, float32_logits, res, g):
    del float32_logits
    axis_size = lax.psum(1, axis_name)
    output, q, k, v, attn_mask, numerator, denominator, max_score = res
    dq = jnp.zeros_like(q, dtype=jnp.float32)
    dk = jnp.zeros_like(k, dtype=jnp.float32)
    dv = jnp.zeros_like(v, dtype=jnp.float32)
    batch, kv_len, num_heads, dim_per_head = k.shape
    scale = jnp.sqrt(q.shape[-1])

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        mask = lax.dynamic_slice_in_dim(attn_mask,
                                        (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1)
        attn_weights = jnp.einsum("bqhd,bkhd->bhqk", q, k) / scale
        # attn_weights = jnp.where(mask, -jnp.inf, attn_weights)
        attn_weights = jnp.where(mask, attn_weights, -jnp.inf)
        exp_weights = jnp.exp(attn_weights - max_score[..., None]) / denominator[..., None]
        ds = jnp.einsum("bqhd,bkhd->bhqk", g, v)
        dl = (ds - jnp.einsum("bqhd,bqhd->bhq", g, output)[..., None]) * exp_weights
        dq = dq + jnp.einsum("bhqk,bkhd->bqhd", dl, k) / scale
        dk = dk + jnp.einsum("bqhd,bhqk->bkhd", q, dl) / scale
        dv = dv + jnp.einsum("bhqk,bqhd->bkhd", exp_weights, g)
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5])
def ring_attention_standard(q, k, v, attn_mask, axis_name, float32_logits=True):
    y, _ = _ring_attention_standard_fwd(q, k, v, attn_mask, axis_name, float32_logits)
    return y


ring_attention_standard.defvjp(_ring_attention_standard_fwd, _ring_attention_standard_bwd)


def _blockwise_attention_fwd(q, k, v, carry, q_chunk_idx_start, k_chunk_idx_start, bias, causal, query_chunk_size,
                             key_chunk_size, deterministic, dropout_rng, attn_pdrop, dtype, policy, precision,
                             prevent_cse):
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    batch, kv_len, num_heads, dim_per_head = v.shape
    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    k = k.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    v = v.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    q, k, v = map(lambda x: jnp.moveaxis(x, 1, 0), (q, k, v))

    numerator, denominator, max_score = carry
    numerator = numerator.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    numerator = jnp.moveaxis(numerator, 1, 0)
    denominator = denominator.reshape((batch, num_heads, num_q, query_chunk_size))
    max_score = max_score.reshape((batch, num_heads, num_q, query_chunk_size))
    denominator, max_score = map(lambda x: rearrange(x, 'b h n c -> n b h c'), (denominator, max_score))

    scale = jnp.sqrt(q.shape[-1])
    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None
    _chunk_bias_fn = partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attn_pdrop, causal, dtype)

    def scan_attention(_, scan):
        q_chunk, numerator_chunk, denominator_chunk, max_score_chunk, q_chunk_idx = scan

        @partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, scan):
            k_chunk, value_chunk, k_chunk_idx = scan
            numerator_chunk, denominator_chunk, prev_max_score_chunk = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bhqk', q_chunk, k_chunk, precision=precision) / scale
            bias_chunk = _chunk_bias_fn(q_chunk_idx_start + q_chunk_idx, k_chunk_idx_start + k_chunk_idx)
            attn_weights = attn_weights + bias_chunk

            max_score_chunk = jnp.maximum(prev_max_score_chunk, jnp.max(attn_weights, axis=-1))
            max_score_chunk = lax.stop_gradient(max_score_chunk)
            exp_weights = jnp.exp(attn_weights - max_score_chunk[..., None])
            exp_values = jnp.einsum('bhqk,bkhd->bqhd', exp_weights, value_chunk, precision=precision)
            correction = rearrange(jnp.exp(prev_max_score_chunk - max_score_chunk), 'b h q -> b q h')[..., None]
            numerator_chunk = numerator_chunk * correction + exp_values
            denominator_chunk = denominator_chunk * jnp.exp(prev_max_score_chunk - max_score_chunk) + exp_weights.sum(
                axis=-1)
            return (numerator_chunk, denominator_chunk, max_score_chunk), None

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, k_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = q_chunk_idx_start + q_chunk_idx < k_chunk_idx_start + k_chunk_idx
            return jax.lax.cond(
                skip_block,
                lambda carry, args: (carry, None),
                scan_kv_block,
                carry,
                args
            )

        (numerator_chunk, denominator_chunk, max_score_chunk), _ = lax.scan(
            skip_upper_half, init=(numerator_chunk, denominator_chunk, max_score_chunk),
            xs=(k, v, jnp.arange(0, num_kv))
        )
        # output_chunk = (numerator_chunk / denominator_chunk).astype(dtype)
        output_chunk = numerator_chunk / rearrange(denominator_chunk, 'b h q -> b q h')[..., None].astype(dtype)
        return (), (output_chunk, numerator_chunk, denominator_chunk, max_score_chunk)

    _, (_, numerator, denominator, max_score) = lax.scan(scan_attention, init=(), xs=(
        q, numerator, denominator, max_score, jnp.arange(0, num_q)))

    numerator = jnp.moveaxis(numerator, 1, 0)
    numerator = numerator.reshape((batch, q_len, num_heads, dim_per_head))
    denominator, max_score = map(lambda x: rearrange(x, 'n b h c -> b h n c'), (denominator, max_score))
    denominator = denominator.reshape((batch, num_heads, q_len))
    max_score = max_score.reshape((batch, num_heads, q_len))

    return numerator, denominator, max_score


def _blockwise_attention_bwd(q, k, v, g, carry, q_chunk_idx_start, k_chunk_idx_start, bias, causal, query_chunk_size,
                             key_chunk_size, deterministic, dropout_rng, attn_pdrop, dtype, policy, precision,
                             prevent_cse):
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    batch, kv_len, num_heads, dim_per_head = v.shape
    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    dq, dk, dv, output, denominator, max_score = carry

    g = g.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    dq = dq.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    dk = dk.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    dv = dv.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    output = output.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    g, dq, dk, dv, output = map(lambda x: jnp.moveaxis(x, 1, 0), (g, dq, dk, dv, output))

    denominator = denominator.reshape((batch, num_heads, num_q, query_chunk_size))
    max_score = max_score.reshape((batch, num_heads, num_q, query_chunk_size))
    denominator, max_score = map(lambda x: rearrange(x, 'b h n c -> n b h c'), (denominator, max_score))

    q = q.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    k = k.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    v = v.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    q, k, v = map(lambda x: jnp.moveaxis(x, 1, 0), (q, k, v))

    scale = jnp.sqrt(q.shape[-1])
    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None
    _chunk_bias_fn = partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attn_pdrop, causal, dtype)

    def scan_attention(carry, scan):
        dk, dv = carry
        q_chunk, dq_chunk, g_chunk, output_chunk, denominator_chunk, max_score_chunk, q_chunk_idx = scan
        dl_part = jnp.einsum("bqhd,bqhd->bhq", g_chunk, output_chunk)[..., None]

        @partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, scan):
            k_chunk, value_chunk, k_chunk_idx = scan
            dq_chunk = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bhqk', q_chunk, k_chunk, precision=precision) / scale
            bias_chunk = _chunk_bias_fn(q_chunk_idx_start + q_chunk_idx, k_chunk_idx_start + k_chunk_idx)
            attn_weights = attn_weights + bias_chunk
            exp_weights = jnp.exp(attn_weights - max_score_chunk[..., None]) / denominator_chunk[..., None]

            ds = jnp.einsum("bqhd,bkhd->bhqk", g_chunk, value_chunk)
            dl = (ds - dl_part) * exp_weights
            dq_chunk = dq_chunk + jnp.einsum("bhqk,bkhd->bqhd", dl, k_chunk) / scale
            dk_chunk = jnp.einsum("bqhd,bhqk->bkhd", q_chunk, dl) / scale
            dv_chunk = jnp.einsum("bhqk,bqhd->bkhd", exp_weights, g_chunk)
            return dq_chunk, (dk_chunk, dv_chunk)

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, k_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = q_chunk_idx_start + q_chunk_idx < k_chunk_idx_start + k_chunk_idx
            return lax.cond(
                skip_block,
                lambda carry, args: (
                    carry, (
                        jnp.zeros((batch, key_chunk_size, num_heads, dim_per_head), dtype=dk.dtype),
                        jnp.zeros((batch, key_chunk_size, num_heads, dim_per_head), dtype=dk.dtype),
                    )
                ),
                scan_kv_block,
                carry,
                args
            )

        dq_chunk, (dk_part, dv_part) = lax.scan(
            skip_upper_half, init=dq_chunk, xs=(k, v, jnp.arange(0, num_kv))
        )
        return (dk + dk_part, dv + dv_part), dq_chunk

    (dk, dv), dq = lax.scan(scan_attention, init=(dk, dv),
                            xs=(q, dq, g, output, denominator, max_score, jnp.arange(0, num_q)))

    dq, dk, dv = map(lambda x: jnp.moveaxis(x, 1, 0), (dq, dk, dv))
    dq = dq.reshape((batch, q_len, num_heads, dim_per_head))
    dk = dk.reshape((batch, kv_len, num_heads, dim_per_head))
    dv = dv.reshape((batch, kv_len, num_heads, dim_per_head))

    return dq, dk, dv


'''
Computing ffn blockwise without materializing the large hidden tensor, training 4x longer sequences than the memory-efficient transformer.
Blockwise parallel transformer https://arxiv.org/abs/2305.19370 Liu et al. 2023
'''


def blockwise_ffn(remat_ffn, inputs, chunk_size, deterministic):
    inputs = rearrange(inputs, 'b (c n) d -> b c n d', c=chunk_size)

    def scan_ffn(remat_ffn, carry, hidden_states):
        outputs = remat_ffn(hidden_states, deterministic=deterministic)
        return carry, outputs

    scan_axis = inputs.ndim - 2
    _, output = nn.scan(
        scan_ffn,
        variable_broadcast="params",
        split_rngs={"params": False, "dropout": True},
        in_axes=scan_axis,
        out_axes=scan_axis,
    )(remat_ffn, None, inputs)
    output = rearrange(output, 'b c n d -> b (c n) d')
    return output


def blockwise_attn(query, key, value, bias, deterministic,
                   dropout_rng, attn_pdrop, causal, query_chunk_size,
                   key_chunk_size, dtype, policy, precision, float32_logits,
                   prevent_cse):
    query = query / jnp.sqrt(query.shape[-1]).astype(dtype)
    if float32_logits:
        query = query.astype(jnp.float32)
        key = key.astype(jnp.float32)

    batch, q_len, num_heads, dim_per_head = query.shape
    batch, kv_len, num_heads, dim_per_head = key.shape
    batch, kv_len, num_heads, dim_per_head = value.shape

    num_q = q_len // query_chunk_size
    num_kv = kv_len // key_chunk_size
    query = query.reshape((batch, num_q, query_chunk_size, num_heads, dim_per_head))
    key = key.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))
    value = value.reshape((batch, num_kv, key_chunk_size, num_heads, dim_per_head))

    query = jnp.moveaxis(query, 1, 0)
    key = jnp.moveaxis(key, 1, 0)
    value = jnp.moveaxis(value, 1, 0)

    if bias is not None:
        for bias_dim, broadcast_dim in zip(bias.shape, (batch, num_heads, q_len, kv_len)):
            assert bias_dim == 1 or bias_dim == broadcast_dim
    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_rng, dropout_rng = jax.random.split(dropout_rng)
        attn_dropout = jax.random.bernoulli(attn_dropout_rng, attn_pdrop, (batch, num_heads, q_len, kv_len))
    else:
        attn_dropout = None

    _chunk_bias_fn = partial(
        _chunk_attention_bias,
        query_chunk_size, key_chunk_size, bias, deterministic,
        attn_dropout, attn_pdrop, causal, dtype)

    def scan_attention(carry, args):
        del carry
        query_chunk, query_chunk_idx = args

        @partial(jax.checkpoint, prevent_cse=prevent_cse, policy=policy)
        def scan_kv_block(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args
            (numerator, denominator, prev_max_score) = carry
            attn_weights = jnp.einsum('bqhd,bkhd->bqhk', query_chunk, key_chunk, precision=precision)
            bias_chunk = _chunk_bias_fn(query_chunk_idx, key_chunk_idx)
            bias_chunk = jnp.moveaxis(bias_chunk, 1, 2)
            attn_weights = attn_weights + bias_chunk

            max_score = jnp.max(attn_weights, axis=-1, keepdims=True)
            max_score = jnp.maximum(prev_max_score, max_score)
            max_score = lax.stop_gradient(max_score)
            exp_weights = jnp.exp(attn_weights - max_score)
            exp_values = jnp.einsum(
                'bqhv,bvhd->bqhd', exp_weights, value_chunk, precision=precision
            )
            correction = jnp.exp(prev_max_score - max_score)
            numerator = numerator * correction + exp_values
            denominator = denominator * correction + exp_weights.sum(axis=-1, keepdims=True)
            return (numerator, denominator, max_score), None

        def skip_upper_half(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args
            skip_block = jnp.array(False)
            if causal:
                skip_block = query_chunk_idx < key_chunk_idx
            return lax.cond(
                skip_block,
                lambda carry, args: (carry, None),
                scan_kv_block,
                carry,
                args,
            )

        init_carry = (
            jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
            jnp.zeros((batch, query_chunk_size, num_heads, dim_per_head), dtype=query.dtype),
            (-jnp.inf) * jnp.ones((batch, query_chunk_size, num_heads, 1), dtype=query.dtype),
        )
        (numerator, denominator, max_score), _ = lax.scan(
            skip_upper_half, init_carry, xs=(key, value, jnp.arange(0, num_kv))
        )
        output = (numerator / denominator).astype(dtype)
        return (), output

    _, output = lax.scan(scan_attention, (), xs=(query, jnp.arange(0, num_q)))
    output = rearrange(output, 'n b c h d -> b (n c) h d')
    return output


def _chunk_attention_bias(query_chunk_size, key_chunk_size,
                          bias, deterministic, attn_dropout, attn_pdrop, causal,
                          dtype, query_chunk_idx, key_chunk_idx):
    query_offset = query_chunk_idx * query_chunk_size
    key_offset = key_chunk_idx * key_chunk_size
    chunk_bias = jnp.zeros((1, 1, 1, 1), dtype=dtype)
    if bias is not None:
        chunk_bias = lax.dynamic_slice(
            bias,
            start_indices=(0, 0, query_offset, key_offset),
            slice_sizes=(*bias.shape[:2], min(bias.shape[-2], query_chunk_size), min(bias.shape[-1], key_chunk_size)),
        )

    if causal:
        query_idx = lax.broadcasted_iota(dtype=jnp.int32, shape=(query_chunk_size, 1), dimension=0)
        key_idx = lax.broadcasted_iota(dtype=jnp.int32, shape=(1, key_chunk_size), dimension=1)
        offset = query_offset - key_offset
        query_idx += offset
        causal_mask_value = (query_idx < key_idx) * jnp.finfo(dtype).min
        chunk_bias += causal_mask_value.reshape(1, 1, *causal_mask_value.shape)

    if not deterministic and attn_pdrop > 0.0:
        attn_dropout_slice = lax.dynamic_slice(
            attn_dropout,
            start_indices=(0, 0, query_offset, key_offset),
            slice_sizes=(
                *attn_dropout.shape[:2],
                min(attn_dropout.shape[-2], query_chunk_size),
                min(attn_dropout.shape[-1], key_chunk_size),
            ),
        )
        chunk_bias += attn_dropout_slice * jnp.finfo(dtype).min
    return chunk_bias.astype(dtype)


def ring_flash_dummy(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    from jax.experimental.pallas.ops.tpu.flash_attention import flash_attention, BlockSizes, mha_reference
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    q, k, v = map(lambda x: rearrange(x, 'b q h d -> b h q d'), [q, k, v])
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    block_sizes = BlockSizes(
        block_q=query_chunk_size,
        block_k_major=key_chunk_size,
        block_k=key_chunk_size,
        block_b=1,
        block_q_major_dkv=query_chunk_size,
        block_k_major_dkv=key_chunk_size,
        block_k_dkv=key_chunk_size,
        block_q_dkv=query_chunk_size,
        block_k_major_dq=key_chunk_size,
        block_k_dq=key_chunk_size,
        block_q_dq=query_chunk_size,
    )

    o = flash_attention(
        q, k, v,
        causal=blockwise_kwargs["causal"],
        block_sizes=block_sizes,
        sm_scale=q.shape[-1] ** -0.5,
    )
    output = rearrange(o.astype(v.dtype), 'b h q d -> b q h d')
    return output


def ring_standard_dummy(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    from jax.experimental.pallas.ops.tpu.flash_attention import flash_attention, BlockSizes, mha_reference
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    q, k, v = map(lambda x: rearrange(x, 'b q h d -> b h q d'), [q, k, v])
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    o = mha_reference(
        q, k, v,
        ab=None,
        causal=blockwise_kwargs["causal"],
        sm_scale=q.shape[-1] ** -0.5,
    )
    output = rearrange(o.astype(v.dtype), 'b h q d -> b q h d')
    return output


def _ring_flash_attention_fwd_tpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    q, k, v = map(lambda x: rearrange(x, 'b q h d -> b h q d'), [q, k, v])
    batch, num_heads, q_len, dim_per_head = q.shape
    batch, num_heads, kv_len, dim_per_head = k.shape
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    o = jnp.zeros((batch, num_heads, q_len, dim_per_head)).astype(q.dtype)
    l = jnp.zeros((batch, num_heads, q_len)).astype(q.dtype)
    m = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(q.dtype)

    axis_size = lax.psum(1, axis_name)
    block_size = q_len  # assumes this function is pre-sharded inside shard_map
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    block_sizes = BlockSizes(
        block_q=query_chunk_size,
        block_k_major=key_chunk_size,
        block_k=key_chunk_size,
        block_b=1,
        block_q_major_dkv=query_chunk_size,
        block_k_major_dkv=key_chunk_size,
        block_k_dkv=key_chunk_size,
        block_q_dkv=query_chunk_size,
        block_k_major_dq=key_chunk_size,
        block_k_dq=key_chunk_size,
        block_q_dq=query_chunk_size,
    )

    scale = q.shape[-1] ** -0.5

    def scan_kv_block(carry, idx):
        o, l, m, k, v = carry
        attn_bias_slice = None
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        o, l, m = _flash_attention_fwd(
            q, k, v,
            carry=(o, l, m),
            q_chunk_idx_start=q_chunk_idx_start,
            k_chunk_idx_start=k_chunk_idx_start,
            ab=attn_bias_slice,
            segment_ids=None,
            save_residuals=False,
            causal=blockwise_kwargs["causal"],
            sm_scale=scale,
            block_sizes=block_sizes,
            debug=False
        )
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]),
                   (k, v))
        return (o, l, m, k, v), None

    (o, l, m, _, _), _ = lax.scan(scan_kv_block,
                                  init=(o, l, m, k, v), xs=jnp.arange(0, axis_size))
    output = rearrange(o.astype(v.dtype), 'b h q d -> b q h d')
    return output, (o, q, k, v, attn_bias, l, m)


def _ring_flash_attention_bwd_tpu(axis_name, float32_logits, blockwise_kwargs, res, g):
    del float32_logits
    o, q, k, v, attn_bias, l, m = res
    batch, num_heads, kv_len, dim_per_head = k.shape
    axis_size = lax.psum(1, axis_name)
    dq = jnp.zeros_like(q, dtype=jnp.float32)
    dk = jnp.zeros_like(k, dtype=jnp.float32)
    dv = jnp.zeros_like(v, dtype=jnp.float32)
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]
    block_size = q.shape[2]  # assumes this function is pre-sharded inside shard_map
    scale = q.shape[-1] ** -0.5

    g = rearrange(g, 'b q h d -> b h q d')

    block_sizes = BlockSizes(
        block_q=query_chunk_size,
        block_k_major=key_chunk_size,
        block_k=key_chunk_size,
        block_b=1,
        block_q_major_dkv=query_chunk_size,
        block_k_major_dkv=key_chunk_size,
        block_k_dkv=key_chunk_size,
        block_q_dkv=query_chunk_size,
        block_k_major_dq=key_chunk_size,
        block_k_dq=key_chunk_size,
        block_q_dq=query_chunk_size,
    )

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        # attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1
        # )
        attn_bias_slice = None  # TODO
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        dq_i, dk_i, dv_i, = _flash_attention_bwd(
            save_residuals=False,
            causal=blockwise_kwargs["causal"],
            sm_scale=scale,
            block_sizes=block_sizes,
            debug=False,
            q_chunk_idx_start=q_chunk_idx_start,
            k_chunk_idx_start=k_chunk_idx_start,
            residuals=(q, k, v, attn_bias_slice, None, o, l, m),
            do=g
        )
        dq += dq_i
        dk += dk_i
        dv += dv_i
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
    dq, dk, dv = map(lambda x: rearrange(x, 'b h q d -> b q h d'), (dq, dk, dv))
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6])
def ring_flash_attention_tpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    """
    The ring_flash_attention_tpu function is a TPU-specific implementation of the ring_flash_attention function.
    It takes in the same arguments as ring_flash_attention, but it also takes in blockwise_kwargs, which are used to
    specify how to partition the input tensors into blocks for parallel processing on TPUs. The blockwise kwargs are:
        - num_blocks: number of blocks to split each dimension into (default 1)
        - dims: list of dimensions that will be split into blocks (default [0])

    :param q: Compute the attention weights
    :param k: Define the key, and the v parameter is used to define the value
    :param v: Compute the attention weights and then used to multiply with the attention weights
    :param attn_bias: Mask out the attention weights for certain positions
    :param axis_name: Specify the axis along which the attention is applied
    :param float32_logits: Determine whether to use float32 or int32 for the logits
    :param blockwise_kwargs: Pass the blockwise_compute function to ring_flash_attention
    :return: A tuple
    
    """
    y, _ = _ring_flash_attention_fwd_tpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs)
    return y


ring_flash_attention_tpu.defvjp(_ring_flash_attention_fwd_tpu, _ring_flash_attention_bwd_tpu)


def ring_flash_dummy_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    """
    The ring_flash_dummy_gpu function is a dummy function that does not actually perform any computation.
    It is used to determine the GPU memory usage of the ring_flash_gpu function, which performs multi-head attention.
    The ring_flash_dummy_gpu function has exactly the same inputs and outputs as ring flash gpu, but it does not perform any computation.

    :param q: Compute the query_chunk_size
    :param k: Calculate the output of the ring_flash_dummy function
    :param v: Compute the output
    :param attn_bias: Mask the attention
    :param axis_name: Specify which axis to split the tensor along
    :param float32_logits: Determine whether to use float32 or float64 for the q and k matrices
    :param blockwise_kwargs: Pass parameters to the ring_flash_dummy_gpu function
    :return: A tuple of (output, none)
    
    """
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    output = mha(q, k, v, None, sm_scale=q.shape[-1] ** -0.5)
    return output


def _ring_flash_attention_fwd_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    """
    The _ring_flash_attention_fwd_gpu function is a GPU-only implementation of the ring flash attention
    forward pass. It takes in query, key, and value tensors (q, k, v), an attention bias tensor (attn_bias),
    and axis name for the blockwise dimension. The function then performs a scan over all blocks in the blockwise
    dimension to compute output values for each query position. This is done by first computing logits using q and k;
    then applying softmax to get probabilities; then multiplying these probabilities with v to get output values.

    :param q: Compute the logits
    :param k: Compute the logits
    :param v: Compute the output, but what is it?
    :param attn_bias: Mask out certain positions in the attention matrix
    :param axis_name: Specify the axis along which to perform ring flash attention
    :param float32_logits: Determine whether to use float32 or bfloat16 for the logits
    :param blockwise_kwargs: Pass in the parameters for the _mha_forward function
    :return: The output of the attention
    
    """
    if float32_logits:
        q, k = q.astype(jnp.float32), k.astype(jnp.float32)
    batch, q_len, num_heads, dim_per_head = q.shape
    batch, kv_len, num_heads, dim_per_head = k.shape
    attn_bias = attn_bias[:, 0, 0]  # (batch, q_len)

    o = jnp.zeros((batch, q_len, num_heads, dim_per_head)).astype(jnp.bfloat16)
    l = jnp.zeros((batch, num_heads, q_len)).astype(jnp.float32)
    m = jnp.full((batch, num_heads, q_len), -jnp.inf).astype(jnp.float32)

    axis_size = lax.psum(1, axis_name)
    block_size = q_len  # assumes this function is pre-sharded inside shard_map
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]

    scale = q.shape[-1] ** -0.5

    def scan_kv_block(carry, idx):
        o, l, m, k, v = carry
        # attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1
        # )
        attn_bias_slice = None  # TODO
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        o, l, m = _mha_forward(
            q, k, v,
            carry=(o.astype(jnp.float32), l, m),
            q_chunk_idx_start=q_chunk_idx_start,  # TODO handle this
            k_chunk_idx_start=k_chunk_idx_start,
            segment_ids=None,
            sm_scale=scale,
            causal=blockwise_kwargs["causal"],
            block_q=query_chunk_size,
            block_k=key_chunk_size,
            backward_pass_impl='triton',  # unused
            num_warps=None,
            num_stages=2,  # unused
            grid=None,
            interpret=False,
            debug=False
        )
        k, v = map(lambda x: lax.ppermute(x, axis_name, perm=[(i, (i + 1) % axis_size) for i in range(axis_size)]),
                   (k, v))
        return (o, l, m, k, v), None

    (o, l, m, _, _), _ = lax.scan(scan_kv_block,
                                  init=(o, l, m, k, v), xs=jnp.arange(0, axis_size))
    output = o.astype(v.dtype)
    return output, (o, q, k, v, attn_bias, l, m)


from jax.experimental.pallas.ops.attention import _mha_forward as _mha_forward_d, _mha_backward as _mha_backward_d, mha


def _ring_flash_attention_bwd_gpu(axis_name, float32_logits, blockwise_kwargs, res, g):
    """
    The _ring_flash_attention_bwd_gpu function is a GPU-specific implementation of the backward pass for
    the ring flash attention mechanism. It takes in the following arguments:

    :param axis_name: Specify the axis that is being sharded
    :param float32_logits: Determine whether to use the float32 or float16 version of the function
    :param blockwise_kwargs: Pass the blockwise_kwargs dictionary to the _ring_flash_attention function
    :param res: Pass in the result of the forward pass
    :param g: Pass the gradient of the output to this function
    :return: A tuple of 4 elements
    
    """
    del float32_logits
    o, q, k, v, attn_bias, l, m = res
    batch, kv_len, num_heads, dim_per_head = k.shape
    axis_size = lax.psum(1, axis_name)
    dq = jnp.zeros_like(q, dtype=q.dtype)
    dk = jnp.zeros_like(k, dtype=k.dtype)
    dv = jnp.zeros_like(v, dtype=v.dtype)
    query_chunk_size = blockwise_kwargs["query_chunk_size"]
    key_chunk_size = blockwise_kwargs["key_chunk_size"]
    block_size = q.shape[1]  # assumes this function is pre-sharded inside shard_map
    scale = q.shape[-1] ** -0.5

    def scan_kv_block(carry, idx):
        dq, dk, dv, k, v = carry
        # attn_bias_slice = lax.dynamic_slice_in_dim(attn_bias,
        #     (lax.axis_index(axis_name) - idx) % axis_size * kv_len, kv_len, axis=-1
        # )
        attn_bias_slice = None  # TODO
        q_block_idx = lax.axis_index(axis_name)
        k_block_idx = (lax.axis_index(axis_name) - idx) % axis_size
        q_chunk_idx_start = q_block_idx * (block_size // query_chunk_size)
        k_chunk_idx_start = k_block_idx * (block_size // key_chunk_size)
        dq_i, dk_i, dv_i = _mha_backward(
            sm_scale=scale,
            causal=blockwise_kwargs["causal"],
            block_q=query_chunk_size,
            block_k=key_chunk_size,
            backward_pass_impl='triton',
            num_warps=None,
            num_stages=2,
            grid=None,
            interpret=False,
            debug=False,
            q_chunk_idx_start=q_chunk_idx_start,
            k_chunk_idx_start=k_chunk_idx_start,
            res=(q, k, v, None, o, l, m),
            do=g,
        )
        dq += dq_i
        dk += dk_i
        dv += dv_i
        k, v, dk, dv = map(lambda x: lax.ppermute(x, axis_name, perm=[(i,
                                                                       (i + 1) % axis_size) for i in range(axis_size)]),
                           (k, v, dk, dv))
        return (dq, dk, dv, k, v), None

    (dq, dk, dv, k, v), _ = lax.scan(scan_kv_block, init=(dq, dk, dv, k, v), xs=jnp.arange(0, axis_size))
    dq, dk, dv = dq.astype(q.dtype), dk.astype(k.dtype), dv.astype(v.dtype)
    return dq, dk, dv, None


@partial(jax.custom_vjp, nondiff_argnums=[4, 5, 6])
def ring_flash_attention_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs):
    """
    The ring_flash_attention_gpu function is a GPU-optimized version of the ring_flash_attention function.
    It uses the same algorithm as ring_flash_attention, but it has been optimized for speed on GPUs.
    The main difference between this function and its CPU counterpart is that this one uses blockwise matrix multiplication to compute attention scores in parallel across multiple blocks of data, rather than computing them sequentially across all data points. This allows us to take advantage of the massive parallelism available on modern GPUs.

    :param q: Query the key-value pairs
    :param k: Determine the number of heads
    :param v: Pass the values to the attention function
    :param attn_bias: Mask the attention weights
    :param axis_name: Specify which axis to perform the attention on
    :param float32_logits: Determine whether to use float32 or float16
    :param blockwise_kwargs: Pass the blockwise_attention function parameters
    :return: The output of the attention layer
    
    """
    y, _ = _ring_flash_attention_fwd_gpu(q, k, v, attn_bias, axis_name, float32_logits, blockwise_kwargs)
    return y


ring_flash_attention_gpu.defvjp(_ring_flash_attention_fwd_gpu, _ring_flash_attention_bwd_gpu)

if __name__ == '__main__':
    def reference_attn(query, key, value, causal, dtype):
        query = query / jnp.sqrt(query.shape[-1]).astype(dtype)
        logits = jnp.einsum("bqhc,bkhc->bhqk", query, key)
        if causal:
            mask_value = jnp.finfo(logits.dtype).min
            _, q_seq_len, _, _ = query.shape
            _, kv_seq_len, _, _ = key.shape
            mask_shape = (q_seq_len, kv_seq_len)
            row_ids = lax.broadcasted_iota(jnp.int32, mask_shape, 0)
            col_ids = lax.broadcasted_iota(jnp.int32, mask_shape, 1)
            causal_mask = (row_ids < col_ids)[None, None, :, :]
            logits = logits + jnp.where(causal_mask, mask_value, 0.0)
        weights = jax.nn.softmax(logits, axis=-1)
        out = jnp.einsum("bhqk,bkhc->bqhc", weights, value)
        return out


    # random inputs
    shape = (1, 32, 8, 64)
    query = jax.random.normal(jax.random.PRNGKey(0), shape)
    key = jax.random.normal(jax.random.PRNGKey(1), shape)
    value = jax.random.normal(jax.random.PRNGKey(2), shape)

    causal = True
    chunk_size = 4
    policy = jax.checkpoint_policies.nothing_saveable()

    blockwise = blockwise_attn(query, key, value, None, False, None, 0.0, causal, chunk_size, chunk_size, jnp.float32,
                               policy, 'float32', True, False)
    reference = reference_attn(query, key, value, causal, 'float32')

    print('max diff sum:', jnp.abs(reference - blockwise).sum())
    print('max diff ele:', jnp.abs(reference - blockwise).max())



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Efficient stochastic rounding implementation."""

from typing import Callable
import jax
import jax.numpy as jnp

NoiseFn = Callable[[tuple[int, ...], jax.Array], jnp.ndarray]


def random_centered_uniform(
        shape: tuple[int, ...], key: jax.Array
) -> jnp.ndarray:
    """Generates uniform number in [-0.5, 0.5]."""
    dtype = jnp.dtype('uint16')
    nbits = jnp.iinfo(dtype).bits

    # Generate random bits.
    bits = jax.random.bits(key, shape, dtype)

    # Align bits with the mantissa of f32.
    nmant = jnp.finfo(jnp.float32).nmant
    r_bitpattern = jnp.uint32(bits) << (nmant - nbits)
    r_bitpattern = r_bitpattern | jnp.float32(1).view(jnp.uint32)
    assert r_bitpattern.dtype == jnp.uint32

    # Gen random floats and shift
    rand_floats = jax.lax.bitcast_convert_type(r_bitpattern, jnp.float32)
    shift = 2 ** (-1 - nbits)
    centered = rand_floats - (1.5 - shift)

    return centered



"""
# FJFormer

a package for custom Jax Flax Functions and Utils
Welcome to fjformer - A collection of useful functions and utilities for Flax and JAX!

## Overview

fjformer is a collection of functions and utilities that can help with various tasks when using Flax and JAX. It
includes
checkpoint savers, partitioning tools, and other helpful functions.
The goal of fjformer is to make your life easier when working with Flax and JAX. Whether you are training a new model,
fine-tuning an existing one, or just exploring the capabilities of these powerful frameworks, fjformer has something to
offer.

## Features

Here are some of the features included in fjformer:

Checkpoint saver: This tool provides an easy way to save and restore checkpoints during training. You can specify how
often to save checkpoints, where to store them, and more.

Partitioning tools: fjformer includes several tools for partitioning data across multiple devices or nodes. These tools
can help you optimize the performance of your models on clusters or distributed systems.

Other utilities: fjformer includes a variety of other helpful functions and utilities and more.

## Getting Started

To get started with fjformer, simply install the package using pip:

```shell
pip install fjformer
```

Once installed, you can import the package and start using its functions and utilities. For example, here's how you can
use the checkpoint saver for loading models like :

```python
from fjformer import StreamingCheckpointer

ckpt = StreamingCheckpointer.load_trainstate_checkpoint('params::<path to model>')

```

or simply getting an optimizer for example adafactor with cosine scheduler :

```python
from jax import numpy as jnp
from fjformer.optimizers import get_adafactor_with_cosine_scheduler

optimizer, scheduler = get_adafactor_with_cosine_scheduler(
    steps=5000,
    learning_rate=5e-5,
    weight_decay=1e-1,
    min_dim_size_to_factor=128,
    decay_rate=0.8,
    decay_offset=0,
    multiply_by_parameter_scale=True,
    clipping_threshold=1.0,
    momentum=None,
    dtype_momentum=jnp.float32,
    weight_decay_rate=None,
    eps=1e-30,
    factored=True,
    weight_decay_mask=None,
)

```

or getting adamw with linear scheduler:

```python
from fjformer.optimizers import get_adamw_with_linear_scheduler

optimizer, scheduler = get_adamw_with_linear_scheduler(
    steps=5000,
    learning_rate_start=5e-5,
    learning_rate_end=1e-5,
    b1=0.9,
    b2=0.999,
    eps=1e-8,
    eps_root=0.0,
    weight_decay=1e-1,
    mu_dtype=None,
)

```

## Documentation

Documentations are available [here](https://erfanzar.github.io/fjformer/docs)

## Contributing

fjformer is an open-source project, and contributions are always welcome! If you have a feature request, bug report, or
just want to help out with development, please check out our GitHub repository and feel free to submit a pull request or
open an issue.

Thank you for using fjformer, and happy training!
"""
from .load import (
    load_and_convert_checkpoint_to_torch,
    float_tensor_to_dtype,
    read_ckpt,
    save_ckpt,
    StreamingCheckpointer,
    get_float_dtype_by_name
)

from .partition_utils import (
    get_jax_mesh,
    names_in_current_mesh,
    get_names_from_partition_spec,
    match_partition_rules,
    flatten_tree,
    get_metrics,
    tree_apply,
    get_weight_decay_mask,
    named_tree_map,
    tree_path_to_string,
    make_shard_and_gather_fns,
    with_sharding_constraint,
    wrap_function_with_rng,
    create_mesh
)

from .monitor import (
    run,
    get_mem,
    is_notebook,
    threaded_log,
    initialise_tracking
)

from .datasets import (
    get_dataloader
)

from .func import (
    transpose,
    global_norm,
    average_metrics
)

from .utils import (
    JaxRNG,
    GenerateRNG,
    init_rng,
    next_rng,
    count_num_params
)

__version__ = '0.0.21'



"""
# File: __init__.py

## Purpose:
This file contains imports for various attention mechanisms used in our project.

## Imports:
1. `efficient_attention` from `efficient_attention` module: This import brings in the `efficient_attention` function,
 which implements an efficient attention mechanism for our project.

2. `dot_product_attention_multihead`, `dot_product_attention_multiquery`, `dot_product_attention_queries_per_head` from
 `flash_attention_0` module: These imports bring in different variations of dot product attention mechanisms for
 multi-head and multi-query scenarios.

3. `ring_attention`, `ring_attention_standard`, `ring_flash_attention_gpu`, `ring_flash_attention_tpu`, `blockwise_ffn`,
 `blockwise_attn` from `flash_attention` module: These imports bring in various attention mechanisms such as
 ring attention, flash attention for GPU and TPU, and blockwise feed-forward network and attention mechanisms.

4. `tpu_flash_attention` `gpu_flash_attention` are imported from jax_flash_attn_tpu/gpu

"""

from .efficient_attention import efficient_attention
from .flash_attention_0 import dot_product_attention_multihead, dot_product_attention_multiquery, \
    dot_product_attention_queries_per_head
from .flash_attention import ring_attention, ring_attention_standard, ring_flash_attention_gpu, \
    ring_flash_attention_tpu, blockwise_ffn, blockwise_attn
from .jax_flash_attn_tpu import flash_attention as tpu_flash_attention, BlockSizes
from .jax_flash_attn_gpu import mha as gpu_flash_attention



# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing fused attention forward and backward pass."""
from __future__ import annotations

import functools
from typing import Any, Optional

import jax
from jax import lax
from jax.experimental import pallas as pl
import jax.numpy as jnp

DEFAULT_MASK_VALUE = -0.7 * float(jnp.finfo(jnp.dtype("float32")).max)


def mha_forward_kernel(
        q_ref,
        k_ref,
        v_ref,
        acc_ref,
        l_ref,
        m_ref,  # Input arrays
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        segment_ids_ref: jax.Array | None,  # segment_id arrays
        o_ref: Any,  # Output
        *residual_refs: Any,  # Residual outputs
        num_heads: int,
        sm_scale: float,
        causal: bool,
        block_q: int,
        block_d: int,
        block_k: int,
):
    """
    The mha_forward_kernel function is the main function that performs the matrix multiplication
    between query and key, as well as between softmax(query*key) and value. It also handles masking
    and residual connections. The kernel is called by mha_forward_kernel_pipeline, which sets up a
    pipeline to run this kernel on multiple blocks of q in parallel.

    :param q_ref: Load the queries from dram to sram
    :param k_ref: Store the key values
    :param v_ref: Load the values from dram
    :param acc_ref: Store the output of the attention layer
    :param l_ref: Store the l_i values
    :param m_ref: Store the max values of qk
    :param q_chunk_idx_start_ref: Store the start index of q_ref
    :param k_chunk_idx_start_ref: Index the k_ref array
    :param segment_ids_ref: jax.Array | None: Determine whether to use segment masking
    :param o_ref: Any: Store the output of the mha_forward_kernel function
    :param # Output
            *residual_refs: Any: Store the residual outputs
    :param # Residual outputs
            num_heads: int: Determine the number of heads in the attention layer
    :param sm_scale: float: Scale the dot product of q and k
    :param causal: bool: Determine whether to use the causal mask or not
    :param block_q: int: Determine the size of the q tile
    :param block_d: int: Specify the size of the output
    :param block_k: int: Determine the size of the kv tile
    :param : Determine the size of the output
    :return: A program
    
    """
    seq_len = q_ref.shape[0]
    start_q = pl.program_id(0)

    # o is the buffer where we accumulate the output on sram.
    # m_i and l_i (see FlashAttention paper) are updated during the k,v loop.
    m_i = pl.load(m_ref, (pl.dslice(start_q * block_q, block_q),))
    l_i = pl.load(l_ref, (pl.dslice(start_q * block_q, block_q),))
    # acc is the buffer where we accumulate the output on sram.
    o = pl.load(acc_ref, (pl.dslice(start_q * block_q, block_q), pl.dslice(None)))

    q_chunk_idx_start = pl.load(q_chunk_idx_start_ref, (0,))
    k_chunk_idx_start = pl.load(k_chunk_idx_start_ref, (0,))

    # Load q: it will stay in L1 throughout. Indices form a matrix because we
    # read, compute, and write all in 2d chunks. 1 element ~= 1 CUDA thread index.
    # q tile has shape [block_q, block_d], block_d == head_dim.
    curr_q_slice = pl.dslice(start_q * block_q, block_q)
    q = pl.load(q_ref, (curr_q_slice, pl.dslice(None)))
    q_segment_ids = (
        None
        if segment_ids_ref is None
        else pl.load(segment_ids_ref, (curr_q_slice,))
    )

    # In FlashAttention algorithm 1 there are 2 loops: slow over tiles of kv (size
    # (Bc == block_k here), and fast over blocks of q (size Br == block_q here).
    # Here we only loop over blocks of kv to process entire seq_len, the loop over
    # blocks of q is carried out by the grid.
    def body(start_k, carry):
        o_prev, m_prev, l_prev = carry
        curr_k_slice = pl.dslice(start_k * block_k, block_k)

        k = pl.load(k_ref, (curr_k_slice, slice(None)))
        kv_segment_ids = (
            None
            if segment_ids_ref is None
            else pl.load(segment_ids_ref, (curr_k_slice,))
        )
        qk = pl.dot(q, k.T)  # [block_q, block_k]
        if sm_scale != 1.:
            qk *= sm_scale  # [block_q, block_k]

        # Avoids Triton crash.
        # if num_heads > 2:
        #   qk = qk.astype(q_ref.dtype)
        #   qk = qk.astype(jnp.float32)

        if causal or segment_ids_ref is not None:
            mask = None
            if segment_ids_ref is not None:
                mask = segment_mask(q_segment_ids, kv_segment_ids)
            if causal:
                span_q = (q_chunk_idx_start + start_q) * block_q + jnp.arange(block_q)
                span_k = (k_chunk_idx_start + start_k) * block_k + jnp.arange(block_k)
                causal_mask = span_q[:, None] >= span_k[None, :]
                mask = (
                    causal_mask if mask is None else jnp.logical_and(mask, causal_mask)
                )
            # Apply mask to qk.
            qk = jnp.where(mask, qk, DEFAULT_MASK_VALUE)

        m_curr = qk.max(axis=-1)
        m_next = jnp.maximum(m_prev, m_curr)
        correction = jnp.exp(m_prev - m_next)
        l_prev_corr = correction * l_prev
        s_curr = jnp.exp(
            qk - m_next[:, None]
        )  # Use m_next instead of m_curr to avoid a correction on l_curr
        l_curr = s_curr.sum(axis=-1)
        l_next = l_prev_corr + l_curr
        l_next_rcp = 1. / l_next
        s_curr = s_curr * l_next_rcp[:, None]
        o_prev_corr = (l_prev_corr * l_next_rcp)[:, None] * o_prev
        v = pl.load(v_ref, (curr_k_slice, pl.dslice(block_d)))
        o_curr = pl.dot(s_curr.astype(v.dtype), v)

        o_next = o_prev_corr + o_curr
        return o_next, m_next, l_next

    if causal:
        # Ceildiv (`pl.cdiv` and `//` do not work due to type of start_q)
        upper_bound = lax.min(
            lax.div(block_q * (q_chunk_idx_start + start_q + 1) + block_k - 1, block_k) - k_chunk_idx_start,
            pl.cdiv(seq_len, block_k))
    else:
        upper_bound = pl.cdiv(seq_len, block_k)  # type: ignore
    o, m_i, l_i = lax.fori_loop(0, upper_bound, body, (o, m_i, l_i))

    if residual_refs:
        l_ref, m_ref = residual_refs
        pl.store(l_ref, (curr_q_slice,), l_i)
        pl.store(m_ref, (curr_q_slice,), m_i)
    # Write output to dram.
    o = o.astype(o_ref.dtype)
    pl.store(o_ref, (curr_q_slice, pl.dslice(None)), o)


def segment_mask(
        q_segment_ids: jax.Array,
        kv_segment_ids: jax.Array,
):
    # [B, T, 1] or [T, 1]
    """
    The segment_mask function is used to mask out the attention scores for
    the query-key pairs that are not in the same segment. This is done by
    creating a boolean array of shape [B, T, S] or [T, S], where B and T are
    the batch size and sequence length of the query tensor respectively; and S
    is the sequence length of key/value tensors. The boolean array has True at
    positions where q_segment_ids == kv_segment_ids (i.e., when they belong to
    the same segment), otherwise False.

    :param q_segment_ids: jax.Array: Create a mask for the query
    :param kv_segment_ids: jax.Array: Create a mask
    :param : Mask the attention weights
    :return: A boolean mask that is true for all the
    
    """
    q_segment_ids = jnp.expand_dims(q_segment_ids, axis=-1)
    # [B, 1, S] or [1, S]
    if kv_segment_ids.ndim == 1:
        kv_segment_ids = jnp.expand_dims(kv_segment_ids, axis=0)
    else:
        kv_segment_ids = jnp.expand_dims(kv_segment_ids, axis=1)
    return jnp.equal(q_segment_ids, kv_segment_ids).astype(jnp.bool_)


def _mha_forward(
        q,
        k,
        v,
        carry,
        q_chunk_idx_start,
        k_chunk_idx_start,
        segment_ids: jax.Array | None,
        sm_scale: float,
        causal: bool,
        block_q: int,
        block_k: int,
        backward_pass_impl: str,
        num_warps: Optional[int],
        num_stages: int,
        grid: Any,
        interpret: bool,
        debug: bool,
):
    """
    The _mha_forward function is a wrapper for the mha_forward_kernel function.
    It takes in the query, key, and value tensors as well as other parameters such
    as segment ids (if applicable), sm scale (scaling factor for softmax), causal
    (whether or not to use causal attention), block q/k (the size of each chunk of
    the sequence length dimension that will be processed by one thread block). It also takes in num warps which is the number of warps per SM on GPU. The grid parameter specifies how many blocks are launched per SM. If it's None then we use heur

    :param q: Store the input query
    :param k: Compute the attention weights
    :param v: Store the output of the attention
    :param carry: Store the output of the previous iteration
    :param q_chunk_idx_start: Keep track of the current chunk
    :param k_chunk_idx_start: Determine the starting index of the key
    :param segment_ids: jax.Array | None: Determine whether the
    :param sm_scale: float: Scale the softmax function
    :param causal: bool: Determine whether the model is causal or not
    :param block_q: int: Specify the number of blocks in the query
    :param block_k: int: Determine the number of blocks in the kernel
    :param backward_pass_impl: str: Choose the implementation of the backward pass
    :param num_warps: Optional[int]: Set the number of warps
    :param num_stages: int: Control the number of stages in the pipeline
    :param grid: Any: Specify the size of the grid
    :param interpret: bool: Print out the kernel code
    :param debug: bool: Print out the shape of each input and output
    :param : Determine the number of heads
    :return: The following:
    
    """
    del backward_pass_impl
    batch_size, seq_len, num_heads, head_dim = q.shape
    block_q = min(block_q, seq_len)
    block_k = min(block_k, seq_len)
    # Heuristics.
    grid_ = grid
    if grid_ is None:
        grid_ = (pl.cdiv(seq_len, block_q), batch_size, num_heads)

    num_warps_ = num_warps
    if num_warps_ is None:
        num_warps_ = 4 if head_dim <= 64 else 8
    kernel = functools.partial(mha_forward_kernel, num_heads=num_heads,
                               sm_scale=sm_scale, causal=causal, block_q=block_q,
                               block_k=block_k, block_d=head_dim)
    out_shape = [
        jax.ShapeDtypeStruct(shape=q.shape, dtype=q.dtype),  # out
        jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len),  # l
                             dtype=jnp.float32),
        jax.ShapeDtypeStruct(shape=(batch_size, num_heads, seq_len),  # m
                             dtype=jnp.float32)
    ]
    in_specs = [
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # q
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # k
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # v
        pl.BlockSpec(
            lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
        ),  # acc
        pl.BlockSpec(
            lambda _, j, k: (j, k, 0), (None, None, seq_len)
        ),  # l
        pl.BlockSpec(
            lambda _, j, k: (j, k, 0), (None, None, seq_len)
        ),  # m
    ]
    in_specs.append(pl.BlockSpec(lambda _, j, k: (0,), (1,)))
    in_specs.append(pl.BlockSpec(lambda _, j, k: (0,), (1,)))
    in_specs.append(
        None  # type: ignore[arg-type]
        if segment_ids is None
        else pl.BlockSpec(lambda _, j, k: (j, 0), (None, seq_len))
    )
    out, l, m = pl.pallas_call(
        kernel,
        grid=grid_,
        in_specs=in_specs,
        out_specs=[
            pl.BlockSpec(
                lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        num_warps=num_warps_,
        num_stages=num_stages,
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="mha_forward",
    )(q, k, v, *carry, q_chunk_idx_start[None], k_chunk_idx_start[None], segment_ids)
    return out, l, m


def _preprocess_backward_kernel(out_ref, dout_ref, l_ref,
                                new_dout_ref, delta_ref, *,
                                block_q: int):
    pid_m = pl.program_id(0)

    off_m = pl.ds(pid_m * block_q, block_q)
    # load
    o = pl.load(out_ref, (off_m, slice(None))).astype(jnp.float32)
    do = pl.load(dout_ref, (off_m, slice(None))).astype(jnp.float32)
    denom = pl.load(l_ref, (off_m,)).astype(jnp.float32)
    # compute
    do = do / denom[:, None]
    delta = jnp.sum(o * do, axis=1)
    # write-back
    pl.store(new_dout_ref, (off_m, slice(None)),
             do.astype(new_dout_ref.dtype))
    pl.store(delta_ref, (off_m,), delta.astype(delta_ref.dtype))


@jax.named_scope("preprocess_backward")
def _preprocess_backward(out, do, l, block_q: int,
                         debug: bool, interpret: bool):
    """
    The _preprocess_backward function is the backward pass of the preprocess function.
    It takes in a tensor ``out``, which is the output of _preprocess_forward, and two other
    tensors ``do`` and ``l``. The first one represents a scaled version of the attention weights
    and has shape (batch_size, seq_len, num_heads). The second one represents an intermediate value used to compute
    the gradient with respect to ln(q) and has shape (batch_size * num heads). It returns two tensors: do' which
    represents a scaled version of datt

    :param out: Compute the do_scaled parameter
    :param do: Store the output of the previous layer
    :param l: Store the length of each sequence in a batch
    :param block_q: int: Control the number of threads in a block
    :param debug: bool: Enable the debug mode of pallas
    :param interpret: bool: Enable the interpreter mode
    :return: A tuple of two arrays
    
    """
    batch_size, seq_len, num_heads, head_dim = out.shape
    out_shape = [
        jax.ShapeDtypeStruct(do.shape, do.dtype),
        jax.ShapeDtypeStruct(l.shape, l.dtype),
    ]
    do_scaled, delta = pl.pallas_call(
        functools.partial(_preprocess_backward_kernel, block_q=block_q),
        grid=(pl.cdiv(seq_len, block_q), batch_size, num_heads),
        in_specs=[
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        out_specs=[
            pl.BlockSpec(lambda _, j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)),
            pl.BlockSpec(lambda _, j, k: (j, k, 0), (None, None, seq_len)),
        ],
        num_warps=4,
        num_stages=3,
        out_shape=out_shape,
        debug=debug,
        interpret=interpret,
        name="mha_preprocess_backward")(out, do, l)
    return do_scaled, delta


def mha_backward_kernel(
        # Inputs
        q_ref,
        k_ref,
        v_ref,
        segment_ids_ref: jax.Array | None,
        out_ref,
        do_scaled_ref,
        l_ref,
        m_ref,
        delta_ref,
        _,
        q_chunk_idx_start_ref,
        k_chunk_idx_start_ref,
        # Outputs
        dq_ref,
        dk_ref,
        dv_ref,
        *,
        sm_scale: float,
        causal: bool,
        block_q: int,
        block_d: int,
        block_k: int,
):
    """
    The mha_backward_kernel function is the backward pass of the multi-head attention
    module. It takes in a reference to q, k, v and out tensors as well as references to
    the segment_ids (if any), do_scaled (which is dout * scaled), l and m tensors. The
    l and m tensors are used for calculating the softmax gradient. The function also takes in a delta value which is used for calculating gradients when there are masked values present in qk or when causal attention has been applied. Finally it also takes in references to chunk indices that indicate where each chunk starts within its respective sequence

    :param # Inputs
            q_ref: Load the query tensor
    :param k_ref: Load the k_ref array
    :param v_ref: Store the value matrix
    :param segment_ids_ref: jax.Array | None: Determine whether or not to use segment masking
    :param out_ref: Store the output of the mha_forward_kernel function
    :param do_scaled_ref: Store the scaled output of the dot product between q and k
    :param l_ref: Store the logits
    :param m_ref: Store the maximum value of qk
    :param delta_ref: Store the difference between the logits and softmax(logits)
    :param _: Pass in the device
    :param q_chunk_idx_start_ref: Keep track of the start index for each chunk
    :param k_chunk_idx_start_ref: Store the index of the current chunk in k_ref
    :param # Outputs
            dq_ref: Store the output of the function
    :param dk_ref: Store the gradients of k_ref
    :param dv_ref: Store the output of the function
    :param *: Pass in the parameters for the mha_backward_kernel function
    :param sm_scale: float: Scale the qk matrix
    :param causal: bool: Determine whether the mask should be applied
    :param block_q: int: Determine the number of q values to be processed at a time
    :param block_d: int: Specify the dimension of the dv_ref array
    :param block_k: int: Determine the size of the block_k
    :param : Determine the number of blocks in the sequence
    :return: None
    
    """
    del out_ref, l_ref  # Not needed
    seq_len = q_ref.shape[0]
    q_chunk_idx_start = pl.load(q_chunk_idx_start_ref, (0,))
    k_chunk_idx_start = pl.load(k_chunk_idx_start_ref, (0,))

    def outer_loop(start_k, _):

        dv = jnp.zeros([block_k, block_d], dtype=jnp.float32)
        dk = jnp.zeros([block_k, block_d], dtype=jnp.float32)
        k = pl.load(k_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
        v = pl.load(v_ref, (pl.ds(start_k * block_k, block_k), slice(None)))
        span_k = (k_chunk_idx_start + start_k) * block_k + jnp.arange(block_k)
        kv_segment_ids = (
            None
            if segment_ids_ref is None
            else pl.load(segment_ids_ref, (pl.ds(start_k * block_k, block_k),))
        )

        def inner_loop(start_q, carry):
            dv, dk = carry
            q = pl.load(q_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
            qk = pl.dot(q, k.T)
            qk = qk.astype(q_ref.dtype)
            qk = qk.astype(jnp.float32)
            if sm_scale != 1.0:
                qk *= sm_scale

            q_segment_ids = (
                None
                if segment_ids_ref is None
                else pl.load(segment_ids_ref, (pl.ds(start_q * block_q, block_q),))
            )

            if causal or segment_ids_ref is not None:
                mask = None
                if segment_ids_ref is not None:
                    mask = segment_mask(q_segment_ids, kv_segment_ids)

                if causal:
                    span_q = (q_chunk_idx_start + start_q) * block_q + jnp.arange(block_q)
                    causal_mask = span_q[:, None] >= span_k[None, :]
                    mask = (
                        causal_mask
                        if mask is None
                        else jnp.logical_and(mask, causal_mask)
                    )
                qk = jnp.where(mask, qk, DEFAULT_MASK_VALUE)

            m = pl.load(m_ref, (pl.ds(start_q * block_q, block_q),))
            p = jnp.exp(qk - m[:, None])
            do = pl.load(do_scaled_ref, (pl.ds(start_q * block_q, block_q), slice(None)))
            dv = dv + pl.dot(p.astype(do.dtype).T, do)
            di = pl.load(delta_ref, (pl.ds(start_q * block_q, block_q),))
            dp = jnp.zeros((block_q, block_k), dtype=jnp.float32) - di[:, None]
            dp = dp + pl.dot(do, v.T)
            ds = p * dp
            if sm_scale != 1.0:
                ds = ds * sm_scale
            dk = dk + pl.dot(ds.astype(q_ref.dtype).T, q)
            dq = pl.load(dq_ref, (pl.ds(start_q * block_q, block_q),
                                  slice(None)), eviction_policy="evict_last")
            dq = dq + pl.dot(ds.astype(k.dtype), k).astype(dq.dtype)
            pl.store(dq_ref, (pl.ds(start_q * block_q, block_q),
                              slice(None)), dq, eviction_policy="evict_last")
            return dv, dk

        if causal:
            lower_bound = lax.max(0, lax.div((k_chunk_idx_start + start_k) * block_k, block_q) - q_chunk_idx_start)
        else:
            lower_bound = 0
        dv, dk = lax.fori_loop(lower_bound, pl.cdiv(seq_len, block_q), inner_loop,
                               (dv, dk))
        pl.store(dv_ref, (pl.ds(start_k * block_k, block_k),
                          slice(None)), dv.astype(dv_ref.dtype))
        pl.store(dk_ref, (pl.ds(start_k * block_k, block_k),
                          slice(None)), dk.astype(dk_ref.dtype))

    lax.fori_loop(0, pl.cdiv(seq_len, block_k), outer_loop, None)


def _mha_backward(sm_scale: float, causal: bool, block_q: int, block_k: int,
                  backward_pass_impl: str, num_warps: Optional[int],
                  num_stages: int, grid: Any, interpret: bool,
                  debug: bool, q_chunk_idx_start, k_chunk_idx_start, res, do):
    """
    The _mha_backward function is a helper function that computes the backward pass of
    the MHA operation. It takes in the following arguments:
        - sm_scale: The scale factor for softmax. This is used to prevent overflow when computing softmax.
        - causal: Whether to use causal attention (i.e., mask future positions). If True, then we will only attend to past positions and ignore future ones (this is useful for language modeling). If False, then we will attend over all positions in the sequence (this is useful for translation).
        - block_q: The size of each chunk along q

    :param sm_scale: float: Scale the softmax output
    :param causal: bool: Determine whether the mask is causal or not
    :param block_q: int: Specify the number of rows in a block
    :param block_k: int: Define the block size of the key matrix
    :param backward_pass_impl: str: Specify whether to use the
    :param num_warps: Optional[int]: Specify the number of warps
    :param num_stages: int: Control the number of stages in the pallas kernel
    :param grid: Any: Pass the grid to the kernel
    :param interpret: bool: Enable the interpreter mode
    :param debug: bool: Print the intermediate results of the kernel
    :param q_chunk_idx_start: Determine the starting index of q in a chunked version of q
    :param k_chunk_idx_start: Indicate the start index of the current chunk
    :param res: Pass the output of the forward pass to the backward function
    :param do: Accumulate the gradient into dq
    :return: The gradient for the query, key and value
    
    """
    del num_warps, num_stages, grid
    q, k, v, segment_ids, out, l, m = res

    if backward_pass_impl == "xla":
        raise Exception("use backward_pass_impl == triton")
    elif backward_pass_impl == "triton":
        batch_size, seq_len, num_heads, head_dim = q.shape
        block_q = min(block_q, seq_len)
        block_k = min(block_k, seq_len)
        do_scaled, delta = _preprocess_backward(out, do, l, block_q, debug, interpret)
        # We accumulate into dq so we need to initialize it to zeros.
        dq = jnp.zeros(q.shape, jnp.float32)
        out_shapes = [
            jax.ShapeDtypeStruct(dq.shape, dq.dtype),
            jax.ShapeDtypeStruct(k.shape, k.dtype),
            jax.ShapeDtypeStruct(v.shape, v.dtype),
        ]

        in_specs = [
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(lambda j, k: (j, k, 0), (None, None, seq_len)),
            pl.BlockSpec(
                lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
            ),
        ]
        if segment_ids is None:
            in_specs.insert(3, None)  # type: ignore[arg-type]
            input_output_aliases = {8: 0}
        else:
            in_specs.insert(3, pl.BlockSpec(lambda j, k: (j, 0), (None, seq_len)))
            input_output_aliases = {9: 0}
        in_specs.append(pl.BlockSpec(lambda j, k: (0,), (1,)))
        in_specs.append(pl.BlockSpec(lambda j, k: (0,), (1,)))
        grid = (batch_size, num_heads)
        # TODO(sharadmv): figure out why num_warps=8 doesn't work!
        num_warps = 8
        dq, dk, dv = pl.pallas_call(
            functools.partial(
                mha_backward_kernel,
                block_q=block_q,
                block_d=head_dim,
                block_k=block_k,
                sm_scale=sm_scale,
                causal=causal,
            ),
            grid=grid,
            out_shape=out_shapes,
            in_specs=in_specs,
            out_specs=[
                pl.BlockSpec(
                    lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
                ),
                pl.BlockSpec(
                    lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
                ),
                pl.BlockSpec(
                    lambda j, k: (j, 0, k, 0), (None, seq_len, None, head_dim)
                ),
            ],
            name="mha_backward",
            debug=debug,
            interpret=interpret,
            num_warps=num_warps,
            num_stages=1,
            input_output_aliases=input_output_aliases,
        )(q, k, v, segment_ids, out, do_scaled, l, m, delta, dq, q_chunk_idx_start[None], k_chunk_idx_start[None])
    else:
        raise ValueError(f"Invalid backward pass implementation: {backward_pass_impl}")
    return dq.astype(q.dtype), dk, dv



# This file is taken from the flaxformer library at
# https://github.com/google/flaxformer

# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An implementation of memory-efficient attention.

Original version published here: https://arxiv.org/abs/2112.05682

Also known as Flash Attention: https://arxiv.org/abs/2205.14135
"""

import functools
from typing import Callable, NamedTuple, Optional, Any

import jax
from jax import lax
from jax import numpy as jnp
from jax import random

Array = jax.Array
PRNGKey = Any
DType = Any


def _causal_bias(
        q_len: int,
        k_len: int,
        offset: Optional[int] = None,
        mask_to_bias_factor: float = 1e6,
) -> Array:
    q_idxs = lax.broadcasted_iota(dtype=jnp.int32, shape=(q_len, 1), dimension=0)
    k_idxs = lax.broadcasted_iota(dtype=jnp.int32, shape=(1, k_len), dimension=1)
    if offset is not None:
        q_idxs += offset
    inverted_mask = q_idxs < k_idxs  # broadcasts to shape (q_len, k_len)
    return inverted_mask * (-1 * mask_to_bias_factor)


def _local_causal_bias(
        q_len: int,
        k_len: int,
        query_offset: int,
        key_offset: int,
) -> Array:
    offset = query_offset - key_offset
    return _causal_bias(q_len, k_len, offset=offset)


class _AttentionSummary(NamedTuple):
    """The summary of the attention over a segment of keys and values."""

    # Sum of the values weighted by the exponentiated scores. Array of shape
    # `[batch, queries, heads, queries_per_head, value_features]`.
    numerator: Array
    # Sum of the exponentiated scores per query. Array of shape
    # `[batch, queries, heads, queries_per_head]`.
    denominator: Array
    # Maximum score encountered per query. Array of shape
    # `[batch, queries, heads, queries_per_head]`.
    max_so_far: Array


def _summarize_chunk(
        query: Array,
        key: Array,
        value: Array,
        current_summary: _AttentionSummary,
        bias: Optional[Array],
        precision=None,
) -> _AttentionSummary:
    """Attention for a segment of queries, keys, and values.

    Args:
        query: An array of shape `[batch, q_length, heads, queries_per_head,
            qk_depth_per_head]`.
        key: An array of shape `[batch, kv_length, heads, qk_depth_per_head]`.
        value: An array of shape `[batch, kv_length, heads, v_depth_per_head]`.
        current_summary: The partially summarized queries so far, before adding the
            summarization of this kv chunk.
        bias: bias for the attention weights. This should be broadcastable to the
            shape `[batch, heads, queries_per_head, q_length, kv_length]` This can be
            used for incorporating causal masks, padding masks, proximity bias, etc.
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.

    Returns:
        The summary for this segment, consisting of sum of the sum of the
        values weighted by their exponentiated attention scores, the exponentiated
        attention scores, and the maximum score of this segment.
    """
    batch, q_len, q_heads, queries_per_head, q_feat = query.shape
    del q_feat
    _, kv_len, _, v_feat = value.shape

    (numerator, denominator, max_so_far) = current_summary

    attn_weights = jnp.einsum(
        'bqhnd,bkhd->bqhnk', query, key, precision=precision
    )

    if bias is not None:
        bias = jnp.moveaxis(bias, -2, 1)  # move sequence length outside
        attn_weights += bias

    previous_max = max_so_far
    assert previous_max.shape == (batch, q_len, q_heads, queries_per_head)
    chunk_maxima = jnp.max(attn_weights, axis=-1)
    assert chunk_maxima.shape == (batch, q_len, q_heads, queries_per_head)
    max_so_far = jnp.maximum(max_so_far, chunk_maxima)
    max_so_far = jax.lax.stop_gradient(max_so_far)
    correction = jnp.exp(previous_max - max_so_far)
    assert correction.shape == (batch, q_len, q_heads, queries_per_head)

    corrected_weights = jnp.exp(
        attn_weights - max_so_far[:, :, :, :, jnp.newaxis]
    )

    assert corrected_weights.shape == (
        batch,
        q_len,
        q_heads,
        queries_per_head,
        kv_len,
    )

    expected_numerator_shape = (batch, q_len, q_heads, queries_per_head, v_feat)
    assert numerator.shape == expected_numerator_shape, (
        f'numerator.shape is {numerator.shape}, but expected'
        f' {expected_numerator_shape}.'
    )

    numerator = numerator * correction[:, :, :, :, jnp.newaxis]
    numerator = numerator + jnp.einsum(
        'bqhnv,bvhf->bqhnf', corrected_weights, value, precision=precision
    )

    assert denominator.shape == (batch, q_len, q_heads, queries_per_head)
    denominator = denominator * correction
    denominator = denominator + corrected_weights.sum(axis=-1)

    return _AttentionSummary(
        numerator,
        denominator,
        max_so_far,
    )


def _memory_efficient_attention(
        query,
        key,
        value,
        bias_fn: Callable[[int, int], Array],
        query_chunk_size: int,
        key_chunk_size: int,
        precision=None,
        dtype=jnp.float32,
        use_extra_logit: bool = False,
        causal_mask: bool = False,
):
    """Computes dot-product multiquery-attention given query, key, and value."""
    batch, num_q, heads, queries_per_head, q_feat = query.shape
    batch, num_kv, heads, k_features = key.shape
    batch, num_kv, heads, v_features = value.shape

    num_q_chunks = num_q // query_chunk_size
    num_kv_chunks = num_kv // key_chunk_size

    query = query.reshape(
        (batch, num_q_chunks, query_chunk_size, heads, queries_per_head, q_feat)
    )
    key = key.reshape((batch, num_kv_chunks, key_chunk_size, heads, k_features))
    value = value.reshape(
        (batch, num_kv_chunks, key_chunk_size, heads, v_features)
    )
    # We move the chunk_idx axis to the front to iterate over it with lax.map.
    query = jnp.moveaxis(query, 1, 0)
    key = jnp.moveaxis(key, 1, 0)
    value = jnp.moveaxis(value, 1, 0)

    # The zero_chunk is the output of _summarize_chunk when the inputs are zeros.
    # We define the zero_chunk outside the loops to prevent the compiler from
    # re-creating these arrays in every loop iteration.
    zero_chunk = _AttentionSummary(
        # numerator
        jnp.zeros(
            (batch, query_chunk_size, heads, queries_per_head, v_features),
            dtype=dtype,
        ),
        # denominator
        jnp.zeros(
            (batch, query_chunk_size, heads, queries_per_head), dtype=dtype
        ),
        # max_so_far
        (-jnp.inf)
        * jnp.ones(
            (batch, query_chunk_size, heads, queries_per_head), dtype=dtype
        ),
    )

    def _query_chunk_attention(args):
        query_chunk, query_chunk_idx = args

        @functools.partial(jax.checkpoint, prevent_cse=False)
        def conditional_summarize_fn(carry, args):
            key_chunk, value_chunk, key_chunk_idx = args

            skip_block = jnp.array(False)
            if causal_mask:
                skip_block = query_chunk_idx < key_chunk_idx

            def cond_fn(query, key, value, carry, key_chunk_idx):
                with jax.named_scope('compute_bias'):
                    chunk_bias = bias_fn(query_chunk_idx, key_chunk_idx)
                return (
                    _summarize_chunk(
                        query, key, value, carry, chunk_bias, precision=precision
                    ),
                    None,
                )

            return jax.lax.cond(
                skip_block,
                lambda a, b, c, carry, d: (carry, None),
                cond_fn,
                query_chunk,
                key_chunk,
                value_chunk,
                carry,
                key_chunk_idx,
            )

        (numerator, denominator, max_so_far), _ = jax.lax.scan(
            conditional_summarize_fn,
            zero_chunk,
            xs=(key, value, jnp.arange(0, num_kv_chunks)),
        )

        if use_extra_logit:
            denominator += jnp.exp(-max_so_far)

        return numerator / denominator[:, :, :, :, jnp.newaxis]

    res = lax.map(_query_chunk_attention, xs=(query, jnp.arange(0, num_q_chunks)))

    expected_res_shape = (
        num_q_chunks,
        batch,
        query_chunk_size,
        heads,
        queries_per_head,
        v_features,
    )
    assert (
            res.shape == expected_res_shape
    ), f'res.shape is {res.shape}, but expected {expected_res_shape}.'
    res = jnp.moveaxis(res, 0, 1)
    return res.reshape(batch, num_q, heads, queries_per_head, value.shape[-1])


def dot_product_attention_queries_per_head(
        query: Array,
        key: Array,
        value: Array,
        bias: Optional[Array] = None,
        broadcast_dropout: bool = True,
        rescale_logits: bool = False,
        dropout_rng: Optional[PRNGKey] = None,
        dropout_rate: float = 0.0,
        enable_dropout: bool = True,
        dtype: DType = jnp.float32,
        precision: Optional[lax.Precision] = None,
        use_extra_logit: bool = False,
        float32_logits: bool = False,
        causal_mask: bool = False,
        query_chunk_size: int = 1024,
        key_chunk_size: int = 2048,
) -> Array:
    """Computes dot-product attention given query, key, and value.

    This is a variant of attention that generalizes both multi-head and
    multi-query attention. It features an extra dimension for the query array,
    that specifies the number of queries per head.

    This function is improved by the memory-efficient attention algorithm
    (https://arxiv.org/abs/2112.05682), which is also called FlashAttention
    (https://arxiv.org/abs/2205.14135).

    Note: query, key, value needn't have any batch dimensions.

    Args:
        query: queries for calculating attention with shape of `[batch..., q_length,
            num_heads, queries_per_head, qk_depth_per_head]`.
        key: keys for calculating attention with shape of `[batch..., kv_length,
            num_heads, qk_depth_per_head]`.
        value: values to be used in attention with shape of `[batch..., kv_length,
            num_heads, v_depth_per_head]`.
        bias: bias for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, queries_per_head, q_length, kv_length]` This
            can be used for incorporating causal masks, padding masks, proximity bias,
            etc.
        broadcast_dropout: bool: use a broadcasted dropout along batch dims.
        rescale_logits: bool. Whether to rescale `query` logits by 1/sqrt(depth_kq).
        dropout_rng: JAX PRNGKey: to be used for dropout
        dropout_rate: dropout rate
        enable_dropout: bool, whether to apply dropout
        dtype: the dtype of the computation (default: float32)
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.
        use_extra_logit: whether to include a virtual extra logit equal to zero.
        float32_logits: bool, if True then compute logits in float32 to avoid
            numerical issues with bfloat16.
        causal_mask: Apply a causal mask. This can be used alternatively or in
            addition to the given bias.
        query_chunk_size: Positive integer to control the size of the query chunks.
        key_chunk_size: Positive integer to control the size of the key chunks.

    Returns:
        Output of shape `[batch..., length, num_heads, queries_per_head,
        v_depth_per_head]`.
    """
    assert (
            key.ndim == value.ndim
    ), f'k, v must have same rank. key: {key.shape}, value: {value.shape}'
    assert (
            query.shape[:-4] == key.shape[:-3] == value.shape[:-3]
    ), f'q, k, v batch dim must match. query: {query.shape}'

    assert key.shape[-3] == value.shape[-3], 'k, v lengths must match.'
    assert query.shape[-1] == key.shape[-1], 'q, k depths must match.'

    # Ensure that we have exactly one batch dimension
    orig_batch_dims = query.shape[:-4]
    query = query.reshape(-1, *query.shape[-4:])
    key = key.reshape(-1, *key.shape[-3:])
    value = value.reshape(-1, *value.shape[-3:])
    if bias is not None:
        bias = bias.reshape(-1, *bias.shape[-4:])

    batch_size, query_length, heads, queries_per_head, _ = query.shape
    _, key_length, _, _ = key.shape

    # TODO: Consider automatic padding to remove this constraint.
    if query_length % query_chunk_size != 0 and query_length > query_chunk_size:
        raise ValueError(
            'Sequence length of the query vector %d needs to be less '
            'than, or a multiple of the query_chunk_size %d.'
            % (query_length, query_chunk_size)
        )
    if key_length % key_chunk_size != 0 and key_length > key_chunk_size:
        raise ValueError(
            'Sequence length of the key/value vector %d needs to be less '
            'than, or a multiple of the key_chunk_size %d.'
            % (key_length, key_chunk_size)
        )

    query_chunk_size = min(query_chunk_size, query_length)
    key_chunk_size = min(key_chunk_size, key_length)

    if bias is not None:
        broadcastable_to = (
            batch_size,
            heads,
            queries_per_head,
            query_length,
            key_length,
        )
        # Check that bias is broadcastable as expected:
        for bias_dim, broadcast_dim in zip(bias.shape, broadcastable_to):
            if bias_dim not in [1, broadcast_dim]:
                raise ValueError(
                    f'Expected bias dimensions {bias.shape} to be broadcastable to'
                    f' {broadcastable_to}.'
                )

    if enable_dropout and dropout_rate > 0.0:
        # Precompute dropout
        drop_shape = [batch_size, heads, queries_per_head, query_length, key_length]
        if broadcast_dropout:
            # We mimick the semantics of T5 and broadcast along the "length" dim.
            drop_shape[-2] = 1  # query_length dim
        precomputed_dropout = random.bernoulli(
            dropout_rng, dropout_rate, drop_shape
        )

    def bias_fn(
            query_chunk_idx: int,
            key_chunk_idx: int,
    ) -> Array:
        query_offset = query_chunk_idx * query_chunk_size
        key_offset = key_chunk_idx * key_chunk_size

        local_bias = jnp.zeros((1, 1, 1, 1, 1))
        if bias is not None:
            # If bias is not broadcasted yet, dynamic slice would fail with full slice
            # size. In this case we keep the bias unbroadcasted.
            slice_q_len = min(bias.shape[-2], query_chunk_size)
            slice_k_len = min(bias.shape[-1], key_chunk_size)
            local_bias = lax.dynamic_slice(
                bias,
                # query_offset and key_offset might be > 1 but bias dims might
                # not yet be broadcasted. We rely on the protection against
                # out-of-bounds array accesses built into dynamic_slice.
                start_indices=(0, 0, 0, query_offset, key_offset),
                slice_sizes=(*bias.shape[:3], slice_q_len, slice_k_len),
            )
        if causal_mask:
            causal = _local_causal_bias(
                query_chunk_size, key_chunk_size, query_offset, key_offset
            )
            # add batch, head, and queries_per_head dims
            local_bias += causal.reshape(1, 1, 1, *causal.shape)
        # We implement dropout as part of the bias, which is additive to the
        # attention scores. In some other implementations it is treated as a
        # multiplicative factor applied to the probabilities after softmax.
        if enable_dropout and dropout_rate > 0.0:
            with jax.named_scope('dropout'):
                # If dropout is not broadcasted yet, we need the collapsed dims.
                slice_q_len = min(precomputed_dropout.shape[-2], query_chunk_size)
                slice_k_len = min(precomputed_dropout.shape[-1], key_chunk_size)
                dropout_slice = lax.dynamic_slice(
                    precomputed_dropout,
                    # query_offset and key_offset might be > 1 but dropout dims might
                    # not yet be broadcasted. We rely on the protection against
                    # out-of-bounds array accesses built into dynamic_slice.
                    start_indices=(0, 0, 0, query_offset, key_offset),
                    slice_sizes=(
                        *precomputed_dropout.shape[:3],
                        slice_q_len,
                        slice_k_len,
                    ),
                )
                local_bias -= dropout_slice * 1e6
        return local_bias

    # NOTE: T5 does not explicitly rescale the attention logits by
    #       1/sqrt(depth_kq)!  This is folded into the initializers of the
    #       linear transformations, which is equivalent under Adafactor.
    if rescale_logits:
        depth = query.shape[-1]
        query = query / jnp.sqrt(depth).astype(dtype)

    # Casting logits and softmax computation for float32 for model stability.
    if float32_logits:
        query = query.astype(jnp.float32)
        key = key.astype(jnp.float32)

    result = _memory_efficient_attention(
        query,
        key,
        value,
        bias_fn,
        query_chunk_size=query_chunk_size,
        key_chunk_size=key_chunk_size,
        precision=precision,
        dtype=dtype,
        use_extra_logit=use_extra_logit,
        causal_mask=causal_mask,
    )
    result = result.reshape(*orig_batch_dims, *result.shape[1:])
    return result


def dot_product_attention_multiquery(
        query: Array,
        key: Array,
        value: Array,
        bias: Optional[Array] = None,
        broadcast_dropout: bool = True,
        rescale_logits: bool = True,
        dropout_rng: Optional[PRNGKey] = None,
        dropout_rate: float = 0.0,
        enable_dropout: bool = True,
        dtype: DType = jnp.float32,
        precision: Optional[lax.Precision] = None,
        use_extra_logit: bool = False,
        float32_logits: bool = True,
        causal_mask: bool = False,
        query_chunk_size: int = 1024,
        key_chunk_size: int = 2048,
) -> Array:
    """Computes dot-product multiquery-attention given query, key, and value.

    This is a variant of the multi-head dot product attention introduced in
    https://arxiv.org/abs/1706.03762 and implemented in `dot_product_attention`.
    In this function, the key and the value have 1 head whereas query has 1 or
    more heads. This variant is called "multi-query" attention.

    This implementation is equivalent to
    `dense_attention.dot_product_attention_multiquery`, but is improved by the
    memory-efficient attention algorithm (https://arxiv.org/abs/2112.05682),
    which is also called FlashAttention (https://arxiv.org/abs/2205.14135).

    Note: query, key, value needn't have any batch dimensions.

    Args:
        query: queries for calculating attention with shape of `[batch..., q_length,
            num_heads, qk_depth_per_head]`.
        key: keys for calculating attention with shape of `[batch..., kv_length,
            qk_depth_per_head]`.
        value: values to be used in attention with shape of `[batch..., kv_length,
            v_depth_per_head]`.
        bias: bias for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, q_length, kv_length]` This can be used for
            incorporating causal masks, padding masks, proximity bias, etc.
        broadcast_dropout: bool: use a broadcasted dropout along batch dims.
        rescale_logits: bool. Whether to rescale `query` logits by 1/sqrt(depth_kq).
        dropout_rng: JAX PRNGKey: to be used for dropout
        dropout_rate: dropout rate
        enable_dropout: bool, whether to apply dropout
        dtype: the dtype of the computation (default: float32)
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.
        use_extra_logit: whether to include a virtual extra logit equal to zero.
        float32_logits: bool, if True then compute logits in float32 to avoid
            numerical issues with bfloat16.
        causal_mask: Apply a causal mask. This can be used alternatively or in
            addition to the given bias.
        query_chunk_size: Positive integer to control the size of the query chunks.
        key_chunk_size: Positive integer to control the size of the key chunks.

    Returns:
        Output of shape `[batch..., length, num_heads, v_depth_per_head]`.
    """
    # num_heads, treat current num_heads as queries_per_head
    query = jnp.expand_dims(query, axis=-3)
    # add num_heads dim
    key = jnp.expand_dims(key, axis=-2)
    # add num_heads dim
    value = jnp.expand_dims(value, axis=-2)
    if bias is not None:
        # add num_heads, treat current num_heads dim as queries_per_head
        bias = jnp.expand_dims(bias, axis=-4)
    result = dot_product_attention_queries_per_head(
        query,
        key,
        value,
        bias,
        broadcast_dropout=broadcast_dropout,
        rescale_logits=rescale_logits,
        dropout_rng=dropout_rng,
        dropout_rate=dropout_rate,
        enable_dropout=enable_dropout,
        dtype=dtype,
        precision=precision,
        use_extra_logit=use_extra_logit,
        float32_logits=float32_logits,
        causal_mask=causal_mask,
        query_chunk_size=query_chunk_size,
        key_chunk_size=key_chunk_size,
    )
    return jnp.squeeze(result, axis=-3)  # remove head dim


def dot_product_attention_multihead(
        query: Array,
        key: Array,
        value: Array,
        bias: Optional[Array] = None,
        broadcast_dropout: bool = True,
        rescale_logits: bool = True,
        dropout_rng: Optional[PRNGKey] = None,
        dropout_rate: float = 0.0,
        enable_dropout: bool = True,
        dtype: DType = jnp.float32,
        precision: Optional[lax.Precision] = None,
        use_extra_logit: bool = False,
        float32_logits: bool = True,
        causal_mask: bool = False,
        query_chunk_size: int = 1024,
        key_chunk_size: int = 2048,
) -> Array:
    """Computes dot-product multi-head attention given query, key, and value.

    This implementation is equivalent to `dense_attention.dot_product_attention`,
    but is more memory-efficient.

    Note: query, key, value needn't have any batch dimensions.

    Args:
        query: queries for calculating attention with shape of `[batch..., q_length,
            num_heads, qk_depth_per_head]`.
        key: keys for calculating attention with shape of `[batch..., kv_length,
            num_heads, qk_depth_per_head]`.
        value: values to be used in attention with shape of `[batch..., kv_length,
            num_heads, v_depth_per_head]`.
        bias: bias for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, q_length, kv_length]` This can be used for
            incorporating causal masks, padding masks, proximity bias, etc.
        broadcast_dropout: bool: use a broadcasted dropout along batch dims.
        rescale_logits: bool. Whether to rescale `query` logits by 1/sqrt(depth_kq).
        dropout_rng: JAX PRNGKey: to be used for dropout
        dropout_rate: dropout rate
        enable_dropout: bool, whether to apply dropout
        dtype: the dtype of the computation (default: float32)
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.
        use_extra_logit: whether to include a virtual extra logit equal to zero.
        float32_logits: bool, if True then compute logits in float32 to avoid
            numerical issues with bfloat16.
        causal_mask: Apply a causal mask. This can be used alternatively or in
            addition to the given bias.
        query_chunk_size: Positive integer to control the size of the query chunks.
        key_chunk_size: Positive integer to control the size of the key chunks.

    Returns:
        Output of shape `[batch..., length, num_heads, v_depth_per_head]`.
    """
    # queries_per_head
    query = jnp.expand_dims(query, axis=-2)
    if bias is not None:
        # add num_heads, treat current num_heads dim as queries_per_head
        bias = jnp.expand_dims(bias, axis=-3)
    result = dot_product_attention_queries_per_head(
        query,
        key,
        value,
        bias,
        broadcast_dropout=broadcast_dropout,
        rescale_logits=rescale_logits,
        dropout_rng=dropout_rng,
        dropout_rate=dropout_rate,
        enable_dropout=enable_dropout,
        dtype=dtype,
        precision=precision,
        use_extra_logit=use_extra_logit,
        float32_logits=float32_logits,
        causal_mask=causal_mask,
        query_chunk_size=query_chunk_size,
        key_chunk_size=key_chunk_size,
    )
    return jnp.squeeze(result, axis=-2)  # remove queries_per_head dim



import os
import msgpack
from flax.serialization import from_bytes, to_bytes, to_state_dict
from flax.traverse_util import flatten_dict

from .streamer import StreamingCheckpointer
from jax import numpy as jnp


def get_float_dtype_by_name(dtype):
    return {
        'bf16': jnp.bfloat16,
        'bfloat16': jnp.bfloat16,
        'fp16': jnp.float16,
        'float16': jnp.float16,
        'fp32': jnp.float32,
        'float32': jnp.float32,
        'fp64': jnp.float64,
        'float64': jnp.float64,
    }[dtype]


def float_tensor_to_dtype(tensor, dtype):
    if dtype is None or dtype == '':
        return tensor
    if isinstance(dtype, str):
        dtype = get_float_dtype_by_name(dtype)
    float_dtypes = (jnp.bfloat16, jnp.float16, jnp.float32, jnp.float64)
    if getattr(tensor, 'dtype', None) in float_dtypes:
        tensor = tensor.astype(dtype)
    return tensor


def load_and_convert_checkpoint_to_torch(path, dtype=jnp.float16, transpose_needed=None,
                                         transpose_not_needed=None, select_params_field: bool = True):
    import torch
    if transpose_needed is None:
        transpose_needed = ["kernel"]
    if transpose_not_needed is None:
        transpose_not_needed = ['none']

    def match_keywords(string, ts, ns):
        for t in ts:
            if t not in string:
                return False
        for n in ns:
            if n in string:
                return False
        return True

    _, flax_params = StreamingCheckpointer.load_trainstate_checkpoint('params::' + path)
    flax_params = flatten_dict(flax_params['params'], sep='.') if select_params_field else flatten_dict(flax_params,
                                                                                                        sep='.')
    torch_params = {}
    for key, tensor in flax_params.items():
        if match_keywords(key, transpose_needed, transpose_not_needed):
            tensor = tensor.T
        tensor = float_tensor_to_dtype(tensor, dtype)
        torch_params[key] = torch.from_numpy(tensor)
    return torch_params


def read_ckpt(path: [str, os.PathLike], shard_fns=None, add_extra_past_fix: list = None):
    tensors = {}
    with open(path, 'rb') as stream:
        unpacker = msgpack.Unpacker(stream, read_size=83886080, max_buffer_size=0)
        for key, value in unpacker:
            if add_extra_past_fix is not None:
                key = add_extra_past_fix + key
            key = tuple(key)
            tensor = from_bytes(None, value)
            if shard_fns is not None:
                tensor = shard_fns[key](tensor)
            tensors[key] = tensor
    return tensors


def save_ckpt(train_state, path, gather_fns=None, float_dtype=None):
    train_state = to_state_dict(train_state)
    packer = msgpack.Packer()
    flatten_train_state = flatten_dict(train_state)
    if gather_fns is not None:
        gather_fns = flatten_dict(to_state_dict(gather_fns))

    with open(path, "wb") as stream:
        for key, value in flatten_train_state.items():
            if gather_fns is not None:
                value = gather_fns[key](value)
            value = float_tensor_to_dtype(value, float_dtype)
            stream.write(packer.pack((key, to_bytes(value))))



import builtins

import jax
from jax import core
from jax import lax
import jax.numpy as jnp
import numpy as np

shape = np.shape
ndim = np.ndim
_max = builtins.max


def matmul(
        a: jnp.ndarray,
        b: jnp.ndarray,
        *,
        precision=None,
        dot_general=lax.dot_general
) -> jnp.ndarray:
    """Quantized jax.numpy.matmul.

    Args:
      a: Left-hand side of the matmul.
      b: Right-hand side of the matmul.
      precision: Indicates precision of a and b.
      dot_general: lax.dot_general by default. To use quantized matmul, the
        wrapper of q_dot_general in which TQs and `train` flag are provided
        should be passed into this function.

    Returns:
      An array containing the result with the same dtype as 'a' and 'b'.
    """
    arraylike = (jax.Array, np.ndarray)
    if not isinstance(a, arraylike) or not isinstance(b, arraylike):
        raise TypeError(f"matmul requires array-like arguments, got {a} and {b}")
    for i, x in enumerate((a, b)):
        if ndim(x) < 1:
            msg = (f"matmul input operand {i} must have ndim at least 1, "
                   f"but it has ndim {ndim(x)}")
            raise ValueError(msg)

    dtype = jnp.result_type(a.dtype, b.dtype)
    a = a.astype(dtype)
    b = b.astype(dtype)

    a_is_mat, b_is_mat = (ndim(a) > 1), (ndim(b) > 1)
    a_batch_dims = shape(a)[:-2] if a_is_mat else ()
    b_batch_dims = shape(b)[:-2] if b_is_mat else ()
    num_batch_dims = _max(len(a_batch_dims), len(b_batch_dims))
    a_batch_dims = (None,) * (num_batch_dims - len(a_batch_dims)) + a_batch_dims
    b_batch_dims = (None,) * (num_batch_dims - len(b_batch_dims)) + b_batch_dims

    # Dimensions to squeeze from the inputs.
    a_squeeze = []
    b_squeeze = []

    # Positions of batch dimensions in squeezed inputs.
    a_batch = []
    b_batch = []

    # Desired index in final output of each kind of dimension, in the order that
    # q_dot_general will emit them.
    idx_batch = []
    idx_a_other = []  # other = non-batch, non-contracting.
    idx_b_other = []
    for i, (ba, bb) in enumerate(zip(a_batch_dims, b_batch_dims)):
        if ba is None:
            idx_b_other.append(i)
        elif bb is None:
            idx_a_other.append(i)
        elif core.symbolic_equal_dim(ba, 1):
            idx_b_other.append(i)
            a_squeeze.append(len(idx_batch) + len(idx_a_other) + len(a_squeeze))
        elif core.symbolic_equal_dim(bb, 1):
            idx_a_other.append(i)
            b_squeeze.append(len(idx_batch) + len(idx_b_other) + len(b_squeeze))
        elif core.symbolic_equal_dim(ba, bb):
            a_batch.append(len(idx_batch) + len(idx_a_other))
            b_batch.append(len(idx_batch) + len(idx_b_other))
            idx_batch.append(i)
        else:
            raise ValueError("Incompatible shapes for matmul arguments: {} and {}"
                             .format(shape(a), shape(b)))

    if a_is_mat:
        idx_a_other.append(num_batch_dims)
    if b_is_mat:
        idx_b_other.append(num_batch_dims + a_is_mat)
    perm = np.argsort(np.concatenate([idx_batch, idx_a_other, idx_b_other]))

    a = lax.squeeze(a, tuple(a_squeeze))
    b = lax.squeeze(b, tuple(b_squeeze))
    out = dot_general(
        a,
        b, (((ndim(a) - 1,), (ndim(b) - 1 - b_is_mat,)), (a_batch, b_batch)),
        precision=precision)
    return lax.transpose(out, perm)


def matmul_true_int8(lhs, rhs):
    """
    The matmul_true_int8 function is a helper function that takes in two int8 arrays and returns the result of their matrix multiplication.

    :param lhs: Specify the left hand side of the matrix multiplication
    :param rhs: Specify the right-hand side of the matrix multiplication
    :return: A matrix with elements of type int32
    
    """
    assert lhs.dtype == jnp.int8
    assert rhs.dtype == jnp.int8
    result = jnp.matmul(lhs, rhs, preferred_element_type=jnp.int32)
    assert result.dtype == jnp.int32
    return result


def quant_int8(x):
    """
    The quant_int8 function takes a floating point number and rounds it to the nearest integer.
    If the rounded value is outside the range [-127, 127], then it clips it to that range.
    Finally, we cast this clipped integer into an int8 type.

    :param x: Pass the input to the function
    :return: The rounded and clipped values
    
    """
    return jnp.clip(jnp.round(x), -127, 127).astype(jnp.int8)


def q_matmul_int8(a, w):
    """
    The q_matmul_int8 function performs a quantized matrix multiplication.

    :param a: Calibrate the input data
    :param w: Store the weights of the neural network
    :return: A float32 array, which is the result of an int8 matrix multiplication
    
    """

    # Calibration. Calibration function is also customizable and injectable.
    a_s = 127 / jnp.max(jnp.abs(a), axis=1, keepdims=True)
    w_s = 127 / jnp.max(jnp.abs(w), axis=0, keepdims=True)

    # int8 matmul with int32 accumulator
    result = matmul_true_int8(quant_int8(a * a_s), quant_int8(w * w_s)) / (a_s * w_s)

    return result



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base abstract class for all numerics."""
import abc
import flax.struct


class QNumerics(flax.struct.PyTreeNode, abc.ABC):
    """Numerics for int8, int4, binary, etc."""

    # TODO(lew): Currently this is a part of API, only because it is used to set
    # it in test. Remove and leave only get_dtype(

    @abc.abstractmethod
    def get_dtype(self):
        pass

    @abc.abstractmethod
    def fwd(self, x, context):
        """Forward pass."""
        pass

    @abc.abstractmethod
    def abs_val_mapped_to(self):
        """The value returned is the end of quantization range.

        It could be the biggest value that can be represented by numerical format
        exactly. E.g. in case of int8, 127 . Or it could be edge of the last bucket.
        Edge in case of int8, 127.5
        """
        pass

    @abc.abstractmethod
    def vjp_fwd(self, x, context):
        pass

    @abc.abstractmethod
    def vjp_bwd(self, res, grad):
        pass



import enum

import chex
import jax.numpy as np
from flax.training import common_utils
from jax.scipy.special import logsumexp
import jax
from jax import numpy as jnp, Array
from typing import Mapping, Optional, Tuple, Union


# Mean Squared Error
def mse(labels, predictions):
    """
    The mse function computes the mean squared error between two arrays.

    :param labels: Pass in the actual values of the data, and predictions is used to pass in our predicted values
    :param predictions: Store the predictions made by the model
    :return: The mean squared error (mse) between the labels and predictions
    
    """
    return np.mean((labels - predictions) ** 2)


# Mean Absolute Error
def mae(labels, predictions):
    """
    The mae function calculates the mean absolute error between two lists of numbers.

    :param labels: Store the actual values of the data
    :param predictions: Store the predictions from the model
    :return: The mean absolute error between the labels and predictions
    
    """
    return np.mean(np.abs(labels - predictions))


# Cross Entropy
def cross_entropy(labels, predictions, ignore_index=None):
    """
    The cross_entropy function computes the cross entropy loss between a set of labels and predictions.

    :param labels: Calculate the cross entropy loss
    :param predictions: Calculate the log_softmax
    :param ignore_index: Mask out the loss from a specific class
    :return: The average cross entropy over the batch
    
    """
    labels = jax.nn.one_hot(labels, predictions.shape[-1])
    if ignore_index is not None:
        mask = np.ones_like(labels)
        mask = np.where(labels == ignore_index, 0, mask)
        labels = labels * mask
        predictions = predictions * mask
    log_softmax = predictions - logsumexp(predictions, axis=-1, keepdims=True)
    return -np.sum(labels * log_softmax) / labels.shape[0]


# Binary Cross Entropy
def binary_cross_entropy(labels, predictions):
    """
    The binary_cross_entropy function computes the binary cross entropy loss between
    the labels and predictions. The function takes two arguments:
        1) labels: a tensor of shape (batch_size, num_classes) containing the ground truth class indices for each example in the batch.
        2) predictions: a tensor of shape (batch_size, num_classes) containing model output probabilities for each example in the batch.

    :param labels: Specify the true class of each example
    :param predictions: Calculate the loss
    :return: The cross-entropy loss for binary classification
    
    """
    labels = jax.nn.one_hot(labels, predictions.shape[-1])
    return -np.mean(labels * np.log(predictions + 1e-8) + (1 - labels) * np.log(1 - predictions + 1e-8))


# Negative Log Likelihood
def nll(labels, predictions):
    """
    The nll function computes the negative log likelihood of a set of predictions.

    :param labels: Calculate the loss
    :param predictions: Calculate the loss
    :return: The negative log likelihood of the predictions
    
    """
    return -np.sum(labels * np.log(predictions + 1e-8))


# L2 Loss
def l2(labels, predictions):
    """
    The l2 function computes the sum of squared differences between labels and predictions.

    :param labels: Calculate the difference between the labels and predictions
    :param predictions: Calculate the difference between the labels and predictions
    :return: The sum of the squared differences between labels and predictions
    
    """
    return np.sum((labels - predictions) ** 2)


# Hinge Loss
def hinge(labels, predictions):
    """
    The hinge function is a loss function used for training classifiers.
    The hinge loss is used for &quot;maximum-margin&quot; classification, most notably
    for support vector machines (SVMs).
    For an intended output t = + 1 or - 1 and a classifier score y, the hinge loss of the prediction y is defined as:

    :param labels: Calculate the hinge loss
    :param predictions: Calculate the hinge loss
    :return: The average loss over the training set
    
    """
    return np.mean(np.maximum(0, 1 - labels * predictions))


# Log-Cosh Loss
def log_cosh(labels, predictions):
    """
    The log_cosh function is a loss function that takes in two arguments: labels and predictions.
    The log_cosh function returns the mean of the logarithm of the hyperbolic cosine of (predictions - labels).
    This loss function is used to measure how well our model performs on data it has not seen before.

    :param labels: Pass the actual values of the data
    :param predictions: Pass the predictions of your model
    :return: The logarithm of the hyperbolic cosine of the prediction error
    
    """

    def cosh(x):
        return (np.exp(x) + np.exp(-x)) / 2

    return np.mean(np.log(cosh(predictions - labels)))


def binary_cross_entropy_onehot(labels, predictions):
    """
    The binary_cross_entropy_onehot function takes in two arguments:
        labels: a 1D array of integers, where each integer is the index of the correct class for that example.
        predictions: a 2D array of floats, where each row represents an example and each column represents a class.

    :param labels: Determine the correct class of each image
    :param predictions: Calculate the loss
    :return: The binary cross entropy loss between the labels and predictions
    
    """
    labels = jax.nn.one_hot(labels, predictions.shape[-1])
    return -np.mean(labels * np.log(predictions + 1e-8) + (1 - labels) * np.log(1 - predictions + 1e-8))


def cross_entropy_onehot(labels, predictions):
    """
    The cross_entropy_onehot function takes two arguments:
    labels - a 1D array of integers in the range [0, num_classes)
    predictions - a 2D array of floats with shape (num_examples, num_classes)
    The function returns the cross entropy loss between labels and predictions. The loss is averaged over all examples.

    :param labels: Compute the one-hot encoding of the labels
    :param predictions: Calculate the log_softmax
    :return: The cross entropy loss
    
    """
    labels = jax.nn.one_hot(labels, predictions.shape[-1])
    log_softmax = predictions - logsumexp(predictions, axis=-1, keepdims=True)
    return -np.sum(labels * log_softmax) / labels.shape[0]


def mse_loss(val, target, valid=None):
    """
    The mse_loss function computes the mean squared error between two arrays.

    :param val: Store the output of the model
    :param target: Calculate the loss
    :param valid: Mask out the loss of certain pixels
    :return: The mean square error of the input and target
    
    """
    if valid is None:
        valid = jnp.ones((*target.shape[:2], 1))
    valid = valid.astype(jnp.float32)
    loss = jnp.mean(
        jnp.where(
            valid > 0.0,
            jnp.square(val - target),
            0.0
        )
    )
    return loss


def cross_entropy_loss_and_accuracy(logits: chex.Array, tokens: chex.Array, valid: chex.Array = None):
    """

    :param logits: Logits
    :param tokens: labels
    :param valid: attention_mask
    :return: loss and accuracy
    """
    if valid is None:
        valid = jnp.ones(tokens.shape[:2])
    valid = valid.astype(jnp.float32)
    valid_text_length = jnp.maximum(jnp.sum(valid, axis=-1), 1e-10)
    logits = logits.astype(jnp.float32)  # for numerical stability
    token_log_prob = jnp.squeeze(
        jnp.take_along_axis(
            jax.nn.log_softmax(logits, axis=-1),
            jnp.expand_dims(tokens, -1),
            axis=-1,
        ),
        -1,
    )
    token_log_prob = jnp.where(valid > 0.0, token_log_prob, jnp.array(0.0))
    loss = -jnp.mean(jnp.sum(token_log_prob, axis=-1) / valid_text_length)
    correct = jnp.where(
        valid > 0.0,
        jnp.argmax(logits, axis=-1) == tokens,
        jnp.array(False)
    )
    accuracy = jnp.mean(jnp.sum(correct, axis=-1) / valid_text_length)
    return loss, accuracy


def fused_cross_entropy_loss_and_accuracy(logits: chex.Array, tokens: chex.Array, valid: chex.Array = None):
    """

    :param logits: Logits
    :param tokens: labels
    :param valid: attention_mask
    :return: loss and accuracy
    """
    if valid is None:
        valid = jnp.ones(tokens.shape[:2])
    valid = valid.astype(jnp.float32)
    valid_text_length = jnp.maximum(jnp.sum(valid, axis=-1), 1e-10)
    logits = logits.astype(jnp.float32)
    log_probs = jax.nn.log_softmax(logits, axis=-1)
    token_log_prob = jnp.squeeze(
        jnp.take_along_axis(
            log_probs,
            jnp.expand_dims(tokens, -1),
            axis=-1,
        ),
        -1,
    )
    token_log_prob = jnp.where(valid > 0.0, token_log_prob, jnp.array(0.0))
    loss = -jnp.mean(jnp.sum(token_log_prob, axis=-1) / valid_text_length)
    correct = jnp.where(
        valid > 0.0,
        jnp.argmax(logits, axis=-1) == tokens,
        jnp.array(False)
    )
    accuracy = jnp.mean(jnp.sum(correct, axis=-1) / valid_text_length)
    return loss, accuracy


@jax.custom_vjp
def cross_entropy_with_logits(logits: jnp.ndarray, targets: jnp.ndarray,
                              z_loss: float) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes cross entropy loss with stable custom gradient.

    Computes a stabilized-gradient version of:
      -jnp.sum(targets * nn.log_softmax(logits), axis=-1)

    If z_loss > 0, then an auxiliary loss equal to z_loss*log(z)^2
    will be added to the cross entropy loss (z = softmax normalization constant).
    The two uses of z_loss are:
    1. To keep the logits from drifting too far from zero, which can cause
       unacceptable roundoff errors in bfloat16.
    2. To encourage the logits to be normalized log-probabilities.

    Args:
      logits: [batch, length, num_classes] float array.
      targets: categorical one-hot targets [batch, length, num_classes] float
        array.
      z_loss: coefficient for auxilliary z-loss loss term.

    Returns:
      tuple with the total loss and the z_loss, both
      float arrays with shape [batch, length].
    """
    logits_sum = jax.scipy.special.logsumexp(logits, axis=-1, keepdims=True)
    log_softmax = logits - logits_sum
    loss = -jnp.sum(targets * log_softmax, axis=-1)
    # Add auxilliary z-loss term.
    log_z = jnp.squeeze(logits_sum, axis=-1)
    total_z_loss = z_loss * jax.lax.square(log_z)
    loss += total_z_loss
    return loss, total_z_loss


def _cross_entropy_with_logits_fwd(
        logits: jnp.ndarray,
        targets: jnp.ndarray,
        z_loss: float = 0.0
) -> Tuple[Tuple[Array, Array], Tuple[Array, Array, float, Array, Array, Array, Array]]:
    """Forward-mode of `cross_entropy_with_logits`."""
    max_logit = logits.max(axis=-1, keepdims=True)
    shifted = logits - max_logit
    exp_shifted = jnp.exp(shifted)
    sum_exp = jnp.sum(exp_shifted, axis=-1, keepdims=True)
    log_softmax = shifted - jnp.log(sum_exp)
    loss = -jnp.sum(targets * log_softmax, axis=-1)
    # Add auxilliary z-loss term.
    log_z = jnp.squeeze(jnp.log(sum_exp) + max_logit, axis=-1)
    total_z_loss = z_loss * jax.lax.square(log_z)
    loss += total_z_loss
    return (loss, total_z_loss), (logits, targets, z_loss, exp_shifted, sum_exp, log_softmax, log_z)


def _cross_entropy_with_logits_bwd(
        res: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray,
        jnp.ndarray, jnp.ndarray], g: Tuple[jnp.ndarray, jnp.ndarray]
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Backward-mode of `cross_entropy_with_logits`."""
    g = g[0]  # Ignore z_loss component as that is only used for logging.
    logits, targets, z_loss, exp_shifted, sum_exp, log_softmax, log_z = res
    # z-loss term adds the (2 * z_loss * log_z) factor.
    deriv = (
            jnp.expand_dims(1 + 2 * z_loss * log_z, -1) * exp_shifted / sum_exp -
            targets)
    g_logits = jnp.expand_dims(g, axis=-1) * deriv
    g_targets = -jnp.expand_dims(g, axis=-1) * log_softmax
    return (jnp.asarray(g_logits,
                        logits.dtype), jnp.asarray(g_targets, targets.dtype),
            jnp.array(0.0))  # sets z-loss coeff gradient to 0


cross_entropy_with_logits.defvjp(_cross_entropy_with_logits_fwd,
                                 _cross_entropy_with_logits_bwd)


def compute_weighted_cross_entropy(
        logits: jnp.ndarray,
        targets: jnp.ndarray,
        weights: Optional[jnp.ndarray] = None,
        label_smoothing: float = 0.0,
        z_loss: float = 0.0,
        loss_normalizing_factor: Optional[float] = None
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Compute weighted cross entropy and entropy for log probs and targets.

    Args:
     logits: [batch, length, num_classes] float array.
     targets: categorical targets [batch, length] int array.
     weights: None or array of shape [batch, length].
     label_smoothing: label smoothing constant, used to determine the on and off
       values.
     z_loss: coefficient for auxiliary z-loss loss term.
     loss_normalizing_factor: Constant to divide loss by. If not specified, loss
       will not be normalized. Intended for backward compatibility with T5-MTF
       training. Should not normally be used.

    Returns:
      Tuple of scalar loss, z_loss, and weight sum.
    """
    if logits.ndim != targets.ndim + 1:
        raise ValueError('Incorrect shapes. Got shape %s logits and %s targets' %
                         (str(logits.shape), str(targets.shape)))
    vocab_size = logits.shape[-1]
    confidence = 1.0 - label_smoothing
    low_confidence = (1.0 - confidence) / (vocab_size - 1)
    normalizing_constant = -(
            confidence * jnp.log(confidence) +
            (vocab_size - 1) * low_confidence * jnp.log(low_confidence + 1e-20))
    soft_targets = common_utils.onehot(
        targets, vocab_size, on_value=confidence, off_value=low_confidence)
    total_loss, total_z_loss = cross_entropy_with_logits(
        logits, soft_targets, z_loss=z_loss)
    total_loss = total_loss - normalizing_constant

    weight_sum = np.prod(targets.shape)
    if weights is not None:
        total_loss = total_loss * weights
        total_z_loss = total_z_loss * weights
        weight_sum = jnp.sum(weights)

    # By default, we do not normalize loss based on anything.
    # We don't normalize based on batch size because the optimizers we use are
    # pretty much scale invariant, so this simplifies things.
    # We don't normalize based on number of non-padding tokens in order to treat
    # each token as equally important regardless of sequence length.
    if loss_normalizing_factor is not None:
        total_loss /= loss_normalizing_factor
        total_z_loss /= loss_normalizing_factor
    return jnp.sum(total_loss), jnp.sum(total_z_loss), weight_sum


@enum.unique
class SpecialLossNormalizingFactor(enum.Enum):
    """Specially calculated loss_normalizing_factors, that are not a constant.

    Attributes:
      NUM_REAL_TARGET_TOKENS: Whether to divide the loss by the number of real
        (non-padding) tokens in the current target batch. If
        'decoder_loss_weights' are specified, it will be the sum of the weights.
        Otherwise it will be the number of non-zero 'decoder_target_tokens'.
      NUM_TOTAL_TARGET_TOKENS: Whether to divide the loss by the total number of
        target tokens, i.e., batch_size * target_seq_length (including padding).
      AVERAGE_PER_SEQUENCE: This will first compute the per-sequence loss
        (averaged over the number of real target tokens in the sequence), and then
        compute the average of that over the sequences. This can be preferable to
        NUM_REAL_TARGET_TOKENS for finetuning, because it will weigh all examples
        equally, regardless of sequence length (which can be especially important
        for multi-task finetuning).
    """
    NUM_REAL_TARGET_TOKENS = 1
    NUM_TOTAL_TARGET_TOKENS = 2
    AVERAGE_PER_SEQUENCE = 3


def convert_special_loss_normalizing_factor_to_enum(
        x: str) -> SpecialLossNormalizingFactor:
    """Converts stringified version of LNF to an enum.

    This is useful because gin dynamic registration does not (currently)
    have support for enum.

    Args:
      x: stringified version of SpecialLossNormalizingFactor enum.

    Returns:
      SpecialLossNormalizingFactor enum instance.
    """
    x = x.upper()
    if x == 'NUM_REAL_TARGET_TOKENS':
        return SpecialLossNormalizingFactor.NUM_REAL_TARGET_TOKENS
    if x == 'NUM_TOTAL_TARGET_TOKENS':
        return SpecialLossNormalizingFactor.NUM_TOTAL_TARGET_TOKENS
    if x == 'AVERAGE_PER_SEQUENCE':
        return SpecialLossNormalizingFactor.AVERAGE_PER_SEQUENCE
    raise ValueError(
        'Could not convert string \"%s\" to SpecialLossNormalizingFactor' % x)


@jax.vmap
def _sum_weights_per_segment(positions: jnp.ndarray, segment_ids: jnp.ndarray,
                             weights: jnp.ndarray) -> jnp.ndarray:
    """Sums weights per packed segment to produce a normalizing vector."""

    # NB: Assumes padding only occurs at the end of a sequence.

    def _repeat_last_nonnegative(xs, reverse=False):
        def fn(prev, x):
            y = jnp.where(x == 0, prev, x)
            return y, y

        return jax.lax.scan(fn, jnp.zeros_like(xs[0]), xs, reverse=reverse)[1]

    # Compute final positions per sequence.
    start_positions = positions == 0
    final_positions = jnp.concatenate([start_positions[1:], jnp.ones(1)])
    # Clear padded positions.
    final_positions *= segment_ids != 0
    # Compute cumulative weights, clearing all but final position per sequence.
    final_cumulative_weights = final_positions * jnp.cumsum(weights)
    # Subtract sequences' final weights from cumulative weights of following ones.
    final_total_weights = jnp.concatenate([
        final_cumulative_weights[0:1],
        jnp.diff(_repeat_last_nonnegative(final_cumulative_weights))
    ])
    # Copy final sequence weight to all positions in sequence.
    normalizer = _repeat_last_nonnegative(final_total_weights, reverse=True)
    return normalizer


def get_loss_normalizing_factor_and_weights(
        loss_normalizing_factor: Optional[Union[float, int, str,
        SpecialLossNormalizingFactor]],
        batch: Mapping[str, jnp.ndarray]):
    """Get the float loss_normalizing_factor and loss weights.

    If loss_normalizing_factor is float or None, this will simply return the
    input loss_normalizing_factor and batch.

    If loss_normalizing_factor is a SpecialLossNormalizingFactor, it will
    return a float loss_normalizing_factor and loss weights corresponding to
    the special LNF. See SpecialLossNormalizingFactor for more details.

    Args:
      loss_normalizing_factor: The input LNF, which may be a float, None, or
        SpecialLossNormalizingFactor (or a stringified SLNF).
      batch: Input data batch.

    Returns:
      Tuple of (output_loss_normalizing_factor, loss_weights).
        'output_loss_normalizing_factor' is a scalar float (Python float
        or jnp float).
        'loss_weights' is the per token loss weight JNP array.
    """

    loss_weights = batch.get('decoder_loss_weights', None)
    if (loss_normalizing_factor is None or
            not isinstance(loss_normalizing_factor,
                           (str, SpecialLossNormalizingFactor))):
        return (loss_normalizing_factor, loss_weights)

    if isinstance(loss_normalizing_factor, str):
        loss_normalizing_factor = convert_special_loss_normalizing_factor_to_enum(
            loss_normalizing_factor)

    # If `loss_weights` are not provided, we assume that the padding id is 0 and
    # that non-padding tokens in the decoder all correspond to the positions
    # where loss should be taken. If more fine-grained behavior (e.g., taking
    # loss on subset of 'decoder_target_tokens') is desired, provide
    # `loss_weights` that account for this.
    if loss_weights is None:
        loss_weights = jnp.asarray(batch['decoder_target_tokens'] > 0, jnp.float32)

    output_normalizing_factor = None
    if (loss_normalizing_factor ==
            SpecialLossNormalizingFactor.NUM_REAL_TARGET_TOKENS):
        output_normalizing_factor = jnp.sum(loss_weights)
    elif (loss_normalizing_factor ==
          SpecialLossNormalizingFactor.NUM_TOTAL_TARGET_TOKENS):
        output_normalizing_factor = np.prod(batch['decoder_target_tokens'].shape)
    elif (loss_normalizing_factor ==
          SpecialLossNormalizingFactor.AVERAGE_PER_SEQUENCE):
        if 'decoder_segment_ids' in batch:  # is packed
            norm_vec = _sum_weights_per_segment(batch['decoder_positions'],
                                                batch['decoder_segment_ids'],
                                                loss_weights)
        else:
            norm_vec = jnp.sum(loss_weights, axis=-1, keepdims=True)
        # Handle divide-by-zero.
        loss_weights = jnp.nan_to_num(
            loss_weights / norm_vec, nan=0, posinf=0, neginf=0)
        output_normalizing_factor = jnp.sum(loss_weights)
    else:
        raise ValueError('Unsupported value of loss_normalizing_factor: %s' %
                         str(loss_normalizing_factor))

    return (output_normalizing_factor, loss_weights)
